# -*- coding: utf-8 -*-

"""
Heuristic for cograph editing.

References
----------
.. [1] Christophe Crespelle.
   Linear-Time Minimal Cograph Editing.
   Preprint published 2019.
"""

import random
import networkx as nx

from tralda.datastructures.Tree import Tree, TreeNode


__author__ = 'David Schaller'


def edit_to_cograph(G, run_number=10):
    """Heuristic algorithm for optimal cograph editing.
    
    Time complexity O(|V|^2).
    
    Parameters
    ----------
    G : nexworkx.Graph
        A graph.
    run_number : Number of editing runs from which the best editing result is
        returned.
    
    Returns
    -------
    Tree
        The cotree, i.e., a Tree instance with inner vertex labels 'series'
        and 'parallel', that corresponds to the best editing result.
    
    References
    ----------
    .. [1] Christophe Crespelle.
       Linear-Time Minimal Cograph Editing.
       Preprint published 2019.
    """
    
    ce = CographEditor(G)
    best_cotree = ce.cograph_edit()
    
    return best_cotree.to_cograph()


class CographEditor:
    """Heuristic for cograph editing running in O(|V|^2)..
    
    References
    ----------
    .. [1] Christophe Crespelle.
       Linear-Time Minimal Cograph Editing.
       Preprint published 2019.
    """
    
    def __init__(self, G):
        """Constructor of CographEditor class.
        
        Parameters
        ----------
        G : networkx.Graph
            A graph.
        """
        
        if not isinstance(G, nx.Graph):
            raise TypeError('not a NetworkX Graph')
        
        self.G = G
        self.V = [v for v in G.nodes()]
        
        self.cotrees = []
        self.costs = []
        
        self.best_T = None
        self.best_cost = float('inf')
        
    
    def cograph_edit(self, run_number=10):
        """Run the cograph editing.
        
        Parameters
        ----------
        run_number : Number of editing runs from which the best editing result is
            returned.
        
        Returns
        -------
        Tree
            The cotree, i.e., a Tree instance with inner vertex labels 'series'
            and 'parallel', that corresponds to the best editing result.
        """
        
        for i in range(run_number):
            
            T = Tree(None)
            self.aux_counter = {}
            already_in_T = set()
            leaf_map = {}
            total_cost = 0
            
            # shuffle the vertex list (beginning from second run)
            if i > 0:
                random.shuffle(self.V)
            
            # build starting tree for the first two vertices
            self._start_tree(T, already_in_T, leaf_map)
            
            # incrementally insert vertices while updating total cost
            if len(self.V) > 2:
                for x in self.V[2:]:
                    cost, x_node = self._insert(x, T, already_in_T, leaf_map)
                    
                    # update the number of leaves on the path to the root
                    current = x_node.parent
                    while current:
                        self.aux_counter[current] += 1
                        current = current.parent
                    
                    leaf_map[x] = x_node
                    already_in_T.add(x)
                    total_cost += cost
                
            self.cotrees.append(T)
            self.costs.append(total_cost)
            
            # update the best cost
            if total_cost < self.best_cost:
                self.best_T = T
                self.best_cost = total_cost
            
            # stop if orginal graph was a cograph, i.e. cost is 0
            if self.best_cost <= 0:
                break
            
        return self.best_T
        
    
    def _start_tree(self, T, already_in_T, leaf_map):
        
        if len(self.V) == 0:
            raise RuntimeError('empty graph in cograph editing')
            return
        
        elif len(self.V) == 1:
            T.root = TreeNode(label=self.V[0])
            return
        
        v1, v2 = self.V[0], self.V[1]
        already_in_T.update([v1, v2])
        
        R = TreeNode(label='series')
        self.aux_counter[R] = 2
        T.root = R
        
        if self.G.has_edge(v1, v2):
            v1_node = TreeNode(label=v1)
            self.aux_counter[v1_node] = 1
            v2_node = TreeNode(label=v2)
            self.aux_counter[v2_node] = 1
            R.add_child(v1_node)
            R.add_child(v2_node)
        else:
            N = TreeNode(label='parallel')
            self.aux_counter[N] = 2
            R.add_child(N)
            v1_node = TreeNode(label=v1)
            self.aux_counter[v1_node] = 1
            v2_node = TreeNode(label=v2)
            self.aux_counter[v2_node] = 1
            N.add_child(v1_node)
            N.add_child(v2_node)
            
        leaf_map[v1] = v1_node
        leaf_map[v2] = v2_node
    
        
    def _insert(self, x, T, already_in_T, leaf_map):
        
        # information for non-hollow nodes
        C_nh = {}                   # node u: list of non-hollow children
        C_nh_number = {}            # node u: number of non-hollow children
        
        Nx_number = {}              # node u: number of neighbors of x in V(u)
        completion_forced = {}      # node u: completion-forced or not
        full = {}                   # node u: full or not (i.e. mixed)
        deletion_forced = {}        # node u: deletion-forced or not
        
        Nx_counter = 0
        
        # ---- FIRST STEP ----
        
        # first traversal (C_nh)
        for v in self.G.neighbors(x):
            if v not in already_in_T:
                continue
            
            Nx_counter += 1
            
            v = leaf_map[v]
            C_nh[v]                = []
            C_nh_number[v]         = 0
            Nx_number[v]           = 1
            completion_forced[v]   = True
            full[v]                = True
            deletion_forced[v]     = False
            
            current = v
            while current.parent:
                if current.parent in C_nh:
                    C_nh[current.parent].append(current)
                    break                               # rest of the path is already done
                else:
                    C_nh[current.parent] = [current]
                current = current.parent                # continue path to root
                
        # all nodes in T are adjacent to x
        if self.aux_counter[T.root] == Nx_counter:
            R = T.root
            x_node = TreeNode(label=x)
            self.aux_counter[x_node] = 1
            R.add_child(x_node)
            leaf_map[x] = x_node
            return 0, x_node                            # cost is 0
        # no nodes in T are adjacent to x
        elif Nx_counter == 0:
            # d(R)=1
            if len(T.root.children) == 1:
                N = T.root.children[0]
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                N.add_child(x_node)
            else:
                R_old = T.root
                R_new = TreeNode(label='series')
                self.aux_counter[R_new] = self.aux_counter[R_old]
                N = TreeNode(label='parallel')
                self.aux_counter[N] = self.aux_counter[R_old]
                R_new.add_child(N)
                N.add_child(R_old)
                T.root = R_new
                
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                N.add_child(x_node)
            return 0, x_node                            # cost is 0
                
        # second traversal, postorder
        # (C_nh_number, Nx_number, completion_forced, mixed, deletion_forced)
        stack = [T.root]
        while stack:
            u = stack.pop()
            if not u.children:
                continue
            elif u not in C_nh_number:
                stack.append(u)
                C_nh_number[u] = len(C_nh[u])
                for nh_child in C_nh[u]:
                    stack.append(nh_child)
            else:
                completion_forced_counter = 0
                deletion_forced_counter = 0
                full_counter = 0
                Nx_number[u] = 0
                for nh_child in C_nh[u]:
                    Nx_number[u] += Nx_number[nh_child]
                    if completion_forced[nh_child]:
                        completion_forced_counter += 1
                    if deletion_forced[nh_child]:
                        deletion_forced_counter += 1
                    if full[nh_child]:
                        full_counter += 1
                full[u] = True if Nx_number[u] == self.aux_counter[u] else False
                
                # u completion-forced?
                if (full[u] or
                    (u.label == 'parallel' and len(u.children) == C_nh_number[u]) or
                    (u.label == 'series' and len(u.children) == completion_forced_counter)):
                    completion_forced[u] = True
                else:
                    completion_forced[u] = False
                    
                # u deletion-forced?
                if ((u.label == 'series' and full_counter == 0) or
                    (u.label == 'parallel' and C_nh_number[u] == deletion_forced_counter)):
                    deletion_forced[u] = True
                else:
                    deletion_forced[u] = False
                    
        # ---- SECOND STEP ----
        cost_above = {}                     # node u: cost above u
        C_mixed = {}                        # node u: mixed children of u
        C_full = {}                         # node u: full children of u
        
        stack = [T.root]
        while stack:        
            u = stack.pop()
            C_mixed[u] = []
            C_full[u] = []
            for nh_child in C_nh[u]:
                if not full[nh_child]:      # only mixed nodes are considered
                    stack.append(nh_child)
                    C_mixed[u].append(nh_child)
                else:
                    C_full[u].append(nh_child)
            v = u.parent
            if not v:                       # root has no parent
                cost_above[u] = 0
            elif u.parent.label == 'parallel':
                cost_above[u] = cost_above[v] + Nx_number[v] - Nx_number[u]
            else:
                cost_above[u] = cost_above[v] + ((self.aux_counter[v] - Nx_number[v]) - 
                                                 (self.aux_counter[u] - Nx_number[u]))
        
        mincost = {}
        
        for u in cost_above.keys():         # contains exactly the mixed nodes
            
            # apply Lemma 7 directly if u has exactly 2 children
            if len(u.children) == 2 and u.label == 'parallel':
                for i in range(2):
                    v, v2 = u.children[i], u.children[-1-i]
                    if v in completion_forced and completion_forced[v]:
                        Nx_v2 = Nx_number[v2] if (v2 in Nx_number) else 0
                        cost = cost_above[u] + (self.aux_counter[v] - Nx_number[v]) + Nx_v2
                        if (u not in mincost) or (cost < mincost[u][0]):
                            mincost[u] = (cost, [v])
            elif len(u.children) == 2 and u.label == 'series':
                for i in range(2):
                    v, v2 = u.children[i], u.children[-1-i]
                    if v not in deletion_forced or deletion_forced[v]:
                        Nx_v = Nx_number[v] if (v in Nx_number) else 0
                        Nx_v2 = Nx_number[v2] if (v2 in Nx_number) else 0
                        cost = cost_above[u] + Nx_v + (self.aux_counter[v2] - Nx_v2)
                        if (u not in mincost) or (cost < mincost[u][0]):
                            mincost[u] = (cost, [v2])
            
            # u has at least 3 children
            elif len(u.children) >= 3:
                red = set()
                blue = set()
                
                if u.label == 'parallel':                                   # u is a parallel node
                    # (step 0)
                    if C_nh_number[u] == 1:                                 # clean child (single non-hollow)
                        v = C_nh[u][0]                                      # v is the single nh child of u
                        if completion_forced[v]:
                            cost = cost_above[u] + (self.aux_counter[v] - Nx_number[v])
                            mincost[u] = (cost, [v])
                        continue                                            # otherwise u is not a minimal
                                                                            # insertion node (Lemma 8)
                    
                    # first step
                    for v in C_mixed[u]:
                        if Nx_number[v] >= self.aux_counter[v] - Nx_number[v]:
                            red.add(v)
                        else:
                            blue.add(v)
                    
                    # second step
                    if C_nh_number[u] == len(u.children) and not blue:
                        current_min, current_min_node = float('inf'), None
                        for v in red:
                            # Nx_number[v] - (self.aux_counter[v] - Nx_number[v])
                            diff = 2 * Nx_number[v] - self.aux_counter[v]
                            if diff < current_min:
                                current_min, current_min_node = diff, v
                        red.remove(current_min_node)
                        blue.add(current_min_node)
                        
                    # third step
                    nb_filled = len(C_full[u]) + len(red)
                    if nb_filled < 2:
                        for i in range(2 - nb_filled):
                            current_min, current_min_node = float('inf'), None
                            for v in blue:
                                # (self.aux_counter[v] - Nx_number[v]) - Nx_number[v]
                                diff = self.aux_counter[v] - 2 * Nx_number[v]
                                if diff < current_min:
                                    current_min, current_min_node = diff, v
                            blue.remove(current_min_node)
                            red.add(current_min_node)
                else:                                                       # u is a series node
                    # (step 0)
                    if len(u.children) - len(C_full[u]) == 1:               # clean child (single non-full)
                        v = None
                        for child in u.children:
                            if (child not in full) or (not full[child]):
                                v = child                                   # v is the single non-full child of u
                                break
                        if (v not in deletion_forced) or (deletion_forced[v]):
                            Nx_v = Nx_number[v] if v in Nx_number else 0
                            cost = cost_above[u] + Nx_v
                            mincost[u] = (cost, C_full[u])
                        continue                                            # otherwise u is not a minimal
                                                                            # insertion node (Lemma 8)
                                                                            
                    # first step
                    for v in C_mixed[u]:
                        if self.aux_counter[v] - Nx_number[v] >= Nx_number[v]:
                            blue.add(v)
                        else:
                            red.add(v)
                    
                    # second step
                    if not C_full[u] and not red:
                        current_min, current_min_node = float('inf'), None
                        for v in blue:
                            # (self.aux_counter[v] - Nx_number[v]) - Nx_number[v]
                            diff = self.aux_counter[v] - 2 * Nx_number[v]
                            if diff < current_min:
                                current_min, current_min_node = diff, v
                        blue.remove(current_min_node)
                        red.add(current_min_node)
                        
                    # third step
                    nb_hollowed = len(u.children) - C_nh_number[u] + len(blue)  # number of hollow or blue children
                    if nb_hollowed < 2:
                        for i in range(2 - nb_hollowed):
                            current_min, current_min_node = float('inf' ), None
                            for v in red:
                                # Nx_number[v] - (self.aux_counter[v] - Nx_number[v])
                                diff = 2 * Nx_number[v] - self.aux_counter[v]
                                if diff < current_min:
                                    current_min, current_min_node = diff, v
                            red.remove(current_min_node)
                            blue.add(current_min_node)
                    
                cost = cost_above[u]
                for v in red:
                    cost += (self.aux_counter[v] - Nx_number[v])
                for v in blue:
                    cost += Nx_number[v]
                    
                mincost[u] = (cost, C_full[u] + list(red))
        
        # determine a minimum cost settling
        insertion_mincost, settling_node = float('inf'), None
        for u in mincost.keys():
            if mincost[u][0] < insertion_mincost:
                insertion_mincost, settling_node = mincost[u][0], u
        
        # insert x into tree
        u = settling_node           # node under which x is inserted
        filled = mincost[u][1]      # children of u to be filled (marked nodes in LinearCographDetector)
        
        # parallel node where one child is to be filled
        if u.label == 'parallel' and len(filled) == 1:
            w = filled[0]
            if w.is_leaf():
                new_node = TreeNode(label='series')  # leaves w and x (x added later)
                self.aux_counter[new_node] = 1
                u.remove_child(w)
                u.add_child(new_node)
                new_node.add_child(w)
                
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                new_node.add_child(x_node)
            else:
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                w.add_child(x_node)
        
        # series node and only one child is not to be filled
        elif (u.label == 'series' and 
              len(u.children) - len(filled) == 1):
            set_A = set(filled)       # auxiliary set
            w = None
            for child in u.children:
                if child not in set_A:
                    w = child
                    break
            if w.is_leaf():
                new_node = TreeNode(label='parallel')  # leaves w and x (x added later)
                self.aux_counter[new_node] = 1
                u.remove_child(w)
                u.add_child(new_node)
                new_node.add_child(w)
                
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                new_node.add_child(x_node)
            else:
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                w.add_child(x_node)
        
        else:
            filled_aux_counter = 0
            y = TreeNode(label=u.label)
            for a in filled:
                u.remove_child(a)
                y.add_child(a)
                filled_aux_counter += self.aux_counter[a]
            self.aux_counter[y] = filled_aux_counter
                
            if u.label == 'parallel':
                new_node = TreeNode(label='series')
                self.aux_counter[new_node] = filled_aux_counter
                u.add_child(new_node)
                
                new_node.add_child(y)
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                new_node.add_child(x_node)
            else:
                par = u.parent
                if par is not None:             # u was the root of T
                    par.remove_child(u)
                    par.add_child(y)
                else:
                    T.root = y                  # y becomes the new root
                
                new_node = TreeNode(label='parallel')
                self.aux_counter[new_node] = 0
                y.add_child(new_node)
                new_node.add_child(u)
                x_node = TreeNode(label=x)
                self.aux_counter[x_node] = 1
                new_node.add_child(x_node)
                
                # update the leaf numbers
                self.aux_counter[y] = self.aux_counter[u]
                self.aux_counter[u] -= filled_aux_counter
                self.aux_counter[new_node] = self.aux_counter[u]

        return insertion_mincost, x_node


# -*- coding: utf-8 -*-

"""
Tree Tools.
"""

from tralda.datastructures.Tree import Tree


__author__ = 'David Schaller'


def assert_leaf_sets_equal(trees):
    """Checks whether trees have the same set of unique leaf labels.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    set
        The common set of leaf labels; or False.
        
    Raises
    ------
    TypeError
        In case of input instances that are not of type Tree.
    RuntimeError
        If the sequence contains empty trees, the tree do not share the same
        set of leaves, or the leaf labels are not unique in one tree.
    AttributeError
        If the label attribute is not set for some leaf.
    """
    
    if not trees:
        raise RuntimeError('empty list of trees')
        
    L = None
        
    for T_i in trees:
        
        if not isinstance(T_i, Tree):
            raise TypeError("not a 'Tree' instance")
        
        L2 = set()
        for v in T_i.leaves():
            if v.label in L2:
                    raise RuntimeError('leaf labels not unique')
            else:
                L2.add(v.label)
            
        if not L:
            L = L2
            if len(L) == 0:
                raise RuntimeError('empty tree in tree list')
        elif L != L2:
            return False
    
    return L


# -*- coding: utf-8 -*-

from tralda.datastructures.Tree import Tree, TreeNode, LCA

from tralda.datastructures.LinkedList import LinkedList, LinkedListNode
from tralda.datastructures.DoublyLinkedList import DLList, DLListNode

from tralda.datastructures.AVLTree import TreeSet, TreeDict

from tralda.datastructures.hdtgraph.DynamicGraph import HDTGraph


# -*- coding: utf-8 -*-

""" 
Implementation of the BuildST algorithm described by Deng & Fernández-Baca.

Builds a supertree from a profile of phylogenetic trees. Uses the algorithm
described by Deng and Fernández-Baca in 2016. The class 'BuildST' accepts a 
list of trees and computes the supertree if it exists.


References:
    - Yun Deng and David Fernández-Baca. Fast Compatibility Testing for
      Rooted Phylogenetic Trees. 27th Annual Symposium on Combinatorial
      Pattern Matching (CPM 2016). DOI: 10.4230/LIPIcs.CPM.2016.12
"""


from tralda.datastructures.Tree import Tree, TreeNode
from tralda.datastructures.DoublyLinkedList import DLList
from tralda.datastructures.hdtgraph.DynamicGraph import HDTGraph


__author__ = 'David Schaller'


def build_st(trees):
    """Supertree construction based on BuildST algorithm.
    
    Note that the graph connecting to trees in the input set whenever they
    share at least one leaf label must be connected. Otherwise, the
    implementation returns False, even though a supertree may exist.
    """
    
    st_builder = BuildST(trees)
    return st_builder.run()


class BuildST:
    """BuildST algorithm.
    
    Note that the graph connecting to trees in the input set whenever they
    share at least one leaf label must be connected. Otherwise, the
    implementation returns False, even though a supertree may exist.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    References
    ----------
    - Yun Deng and David Fernández-Baca. Fast Compatibility Testing for
      Rooted Phylogenetic Trees. 27th Annual Symposium on Combinatorial
      Pattern Matching (CPM 2016). DOI: 10.4230/LIPIcs.CPM.2016.12
    """
    
    def __init__(self, trees):
        """Constructor for BuildST algorithm."""
        
        self.trees = trees
        self.Xp = {}                                    # leaf label --> XpNode
        self.marked = set()                             # set of marked nodes
        self.node_to_tree_index = {}
        self.list_pointer = {}                          # node --> DLL element
        self.singleton_pointer = {}                     # node --> DLL element
        self.hdt_graph = HDTGraph()
        
        self.fail_message = ''
        
    
    def run(self):
        """Build the supertree from the given tree list if existent."""
        
        self._prepare_trees()        # tree indices, Xp nodes

        U_init = self._initialize()
        if not U_init:
            return False
        root = self._buildst(U_init)
        if not root:
            return False
        else:
            return Tree(root)
    
    
    def _prepare_trees(self):
        
        for i in range(len(self.trees)):
            tree = self.trees[i]
            for node in tree.preorder():
                # access to tree index i in [k]
                self.node_to_tree_index[node] = i
                if not node.children:
                    if node.label not in self.Xp:
                        self.Xp[node.label] = XpNode(node.label)
                    self.Xp[node.label].leafnodes.append(node)
    
    
    def _initialize(self):
        
        Y = _ConnectedComp(len(self.trees), self, count=len(self.Xp))
        
        for i, tree in enumerate(self.trees):
            
            # Y_init.singleton is the set [k]
            self.singleton_pointer[tree.root] = Y.singleton.append(i)
            
            # append the root to DLList at Y.List[i]
            self.list_pointer[tree.root] = Y.List[i].append(tree.root)
            
            Y.initialize_tree_edges(tree)
            
            # mark root as part of set U_init
            self.marked.add(tree.root)
                          
        # glue together the leafnodes and the labelnodes in Xp
        for labelnode in self.Xp.values():
            for leafnode in labelnode.leafnodes:
                Y.add_edge(labelnode, leafnode)
                
        if self.hdt_graph:
            ett = self.hdt_graph.is_connected()
            if ett:
                Y.representative = self.trees[0].root
            else:
                self.failmessage = 'initialization failed because graph is '\
                                   'not fully connected'
                return False
            
        return Y
    
    
    def _buildst(self, U):
        
        # --------------------------------------------------
        # if |L(U)| = 1 then
        #    return the tree consisting of node r_U,
        #    labeled by the single species in L(U)
        # --------------------------------------------------
        if U.count == 1:
            labelnode = U.get_labelnodes()[0]
            return TreeNode(label=labelnode.label)
        
        r_U = TreeNode()      # create a node r_U
        
        # --------------------------------------------------
        # if |L(U)| = 2 then
        #    return the tree consisting of node r_U and two children
        #    each labeled by a different species in L(U)
        # --------------------------------------------------
        if U.count == 2:
            labels = U.get_labelnodes()
            if len(labels) != 2:
                self.failmessage = 'could not find 2 labeled nodes'
                return
            node1 = TreeNode(label=labels[0].label)
            node2 = TreeNode(label=labels[1].label)
            r_U.add_child(node1)
            r_U.add_child(node2)
            return r_U
        
        # --------------------------------------------------
        # foreach i in [k] such that |U(i)| = 1 do
        #    Let v be the single element in U(i)
        #    U = (U \ {v}) union Ch(v)
        # --------------------------------------------------
        W = [U]                             # list of conn. components
        J = [j for j in U.singleton]        # elements in singleton
        V = [U.List[i][0] for i in J]       # singleton nodes to be deleted
        
        for i in J:
            v_i = U.List[i][0]
            U.List[i].remove_node(self.list_pointer[v_i])
            U.singleton.remove_node(self.singleton_pointer[v_i])
            
            self.marked.discard(v_i)        # unmark v_i and
            for child in v_i.children:      # mark its children
                self.marked.add(child)
                self.list_pointer[child] = U.List[i].append(child)
        
        for v_i in V:
            Y_current = None
            
            # which component in W contains v_i?
            for Y in W:
                if Y.contains_key(v_i):
                    Y_current = Y
                    break
                
            if not Y_current:
                # print("Could not find connected component for v_i", v_i)
                return
            
            for u in v_i.children:
                conn_comp = Y_current.delete_edge(v_i, u)
                if len(conn_comp) == 2:
                    if v_i in conn_comp[0]:
                        
                        # (case 1.1) u is the last child of v_i
                        #            then conn_comp[0] is actually empty
                        if len(conn_comp[0]) == 1:
                            Y_current.set_component(component=conn_comp[1],
                                                    representative=u)
                            break
                        
                        # (case 1.2) two actual components emerge
                        else:
                            Y_current.set_component(component=conn_comp[0],
                                                    representative=v_i)
                            Y_new = _ConnectedComp(Y_current.k, self)
                            Y_new.set_component(component=conn_comp[1], 
                                                representative=u)
                            current_smaller = True \
                                if len(conn_comp[0]) < len(conn_comp[1]) \
                                else False
                    
                    elif v_i in conn_comp[1]:
                        
                        # (case 2.1) u is the last child of v_i
                        #            then conn_comp[1] is actually empty
                        if len(conn_comp[1]) == 1:
                            Y_current.set_component(component=conn_comp[0],
                                                    representative=u)
                            break
                        
                        # (case 2.2) two actual components emerge
                        else:
                            Y_current.set_component(component=conn_comp[1],
                                                    representative=v_i)
                            Y_new = _ConnectedComp(Y_current.k, self)
                            Y_new.set_component(component=conn_comp[0],
                                                representative=u)
                            current_smaller = True \
                                if len(conn_comp[1]) < len(conn_comp[0]) \
                                else False
                    
                    if current_smaller:
                        self._update_conn_comps(Y_current, Y_new,
                                                first_swap=True)
                    else:
                        self._update_conn_comps(Y_new, Y_current)
                    
                    W.append(Y_new)
                    
        # --------------------------------------------------
        # Let W_1, W_2, ..., W_p be the connected components
        # if p = 1 then
        #    return incompatible
        # --------------------------------------------------
        if len(W) == 1:
            self.failmessage = 'the trees are incompatible'
            return False
        
        # --------------------------------------------------
        # foreach j in [p] do
        #    Let t_j = BuildST(W_j)
        #    if t j is a tree then
        #       Add t_j to the set of subtrees of r_U
        #    else
        #       return incompatible
        # --------------------------------------------------
        for W_j in W:
            t_j = self._buildst(W_j)
            if t_j:
                r_U.add_child(t_j)
            else:
                return False
        
        # return the tree with root r_U
        return r_U
    
    
    def _update_conn_comps(self, Y1, Y2, first_swap=False):
        """Updates Y.count, Y.singleton and Y.List of two newly split components.
        
        Component Y1 is supposed to be the smaller one and is scanned for label
        nodes (Xp) and for marked nodes.
        Keyword argument:
            first_swap - if Y1 (smaller component) contains the information
                         about count, List and singleton instead of Y2
                         these attributes are swapped
        """
        
        if first_swap:
            Y1.count, Y2.count = Y2.count, Y1.count
            Y1.List, Y2.List = Y2.List, Y1.List
            Y1.singleton, Y2.singleton = Y2.singleton, Y1.singleton
            
        for v in Y1.keys():
            if isinstance(v, XpNode):
                Y1.count += 1
                Y2.count -= 1
            elif v in self.marked:
                i = self.node_to_tree_index[v]
                Y2.List[i].remove_node(self.list_pointer[v])
                self.list_pointer[v] = Y1.List[i].append(v)
                
                # v becomes singleton
                if len(Y2.List[i]) == 1:
                    singleton_node = Y2.List[i][0]
                    self.singleton_pointer[singleton_node] = Y2.singleton.append(i)
                # index i is no longer singleton, i.e. it must have been a singleton
                elif len(Y2.List[i]) == 0:
                    Y2.singleton.remove_node(self.singleton_pointer[v])
                
                # v becomes singleton
                if len(Y1.List[i]) == 1:                                        
                    self.singleton_pointer[v] = Y1.singleton.append(i)
                # index i is no longer singleton
                elif len(Y1.List[i]) == 2:
                    Y1.singleton.remove_node(self.singleton_pointer[Y1.List[i][0]])


class XpNode:
    """Special type of node for the set Xp (one for each leaf label).""" 
    
    def __init__(self, label):
        
        self.label = label
        self.leafnodes = []         # corresponding treenodes of the profile
    
    def __repr__(self):
        
        return '<XpNodeID:{}, {}>'.format(id(self), self.label)
    
    def __str__(self):
        
        return str(self.label)
        
    
class _ConnectedComp:
    """Connected component for D. & F.-B. algorithm (HDT data structure)."""
    
    def __init__(self, k, buildst, count=0, singleton=None, List=None):
        
        self.k = k
        self.buildst = buildst
        self.hdt_graph = buildst.hdt_graph
        self.count = count
        self.singleton = singleton if singleton else DLList()
        self.List = List if List else [DLList() for i in range(k)]
        
        # to find the component in the HDT datastructure
        self.representative = None
    
    
    def keys(self):
        """Generator for the elements in the component.
        
        Caution: If the HDT datastructure is used this will only work
        if a representative of the component is set correctly.
        """
        
        yield from self.hdt_graph.component_iterator(self.representative)
    
    
    def contains_key(self, key):
        """Check if a given key is element of the component."""
        
        if not self.representative: 
            raise RuntimeError('missing representative for the component (2)')
            
        return self.hdt_graph.connected(self.representative, key)
    
    
    def initialize_tree_edges(self, tree):
        """Add all edges of a tree to the graph datastructure."""
        
        self.hdt_graph.add_loose_tree(tree)
    
    
    def add_edge(self, u, v):
        """Add an edge to the graph datastructure."""
        
        self.hdt_graph.insert_edge(u, v)
    
    
    def delete_edge(self, u, v):
        """Delete an edge and return the 1 or 2 conn. components of u and v."""
        
        self.hdt_graph.delete_edge(u, v)
        ett1 = self.hdt_graph.get_component(u)
        ett2 = self.hdt_graph.get_component(v)
        if ett1 is ett2:
            return [ett1]
        else:
            return [ett1, ett2]
    
    
    def set_component(self, component=None, representative=None):
        """Set the component (self.representative)."""
        
        self.representative = representative
    
    
    def get_labelnodes(self):
        """Return the labelnodes in the connected component."""
        
        result = []
        if not self.representative:
            raise RuntimeError('missing representative for the component (1)')
            
        for node in self.hdt_graph.component_iterator(self.representative):
            if isinstance(node, XpNode):
                result.append(node)
                
        return result



# -*- coding: utf-8 -*-

"""
AVL-tree implementation.
    
Balanced binary search tree implementation of a set (TreeSet) and
dictionary (TreeDict).
"""


__author__ = 'David Schaller'


class TreeSetNode:
    
    __slots__ = ('key', 'parent', 'left', 'right',
                 'height', 'size', 'balance')
    
    def __init__(self, key):
        
        self.key = key
        
        self.parent = None
        self.left = None
        self.right = None
        
        self.height = 1     # height of the subtree
        self.size = 1       # stores number of elements in its subtree
        self.balance = 0
   
    
    def update(self):
        """Update height, size and balance of (the subtree under) the node."""
        
        if self.left and self.right:
            self.height = 1 + max(self.left.height, self.right.height)
            self.size = 1 + self.left.size + self.right.size
            self.balance = self.left.height - self.right.height
        elif self.left:
            self.height = 1 + self.left.height
            self.size = 1 + self.left.size
            self.balance = self.left.height
        elif self.right:
            self.height = 1 + self.right.height
            self.size = 1 + self.right.size
            self.balance = -(self.right.height)
        else:
            self.height = 1
            self.size = 1
            self.balance = 0
        
        return self.balance
    
    
    def left_size(self):
        
        return self.left.size if self.left else 0
    
    
    def right_size(self):
        
        return self.right.size if self.right else 0
    
    
    def __str__(self):
        
        return '<AVLNode: {}>'.format(self.key)
    
    
    def copy(self):
        
        copy = TreeSetNode(self.key)
        copy.height = self.height
        copy.size = self.size
        copy.balance = self.balance
        return copy
 

class TreeSet:
    """AVL tree."""
    
    __slots__ = ('root')
    
    def __init__(self):
        
        self.root = None
    
    
    def __iter__(self):
        
        return TreeSetIterator(self)


    def __next__(self):
        
        pass
            
            
    def __nonzero__(self):
        
        return True if self.root else False
      
    
    def __len__(self):
        
        return self.root.size if self.root else 0
    
    
    def __contains__(self, item):
        
        return self._find(item) is not None
    
    
    def __getitem__(self, index):
        """Return the element at the index.
        
        Same as 'key_at(index)'."""
        
        return self._node_at(index).key
    
    
    def add(self, item):
        """Insert an item."""
        
        if not self.root:
            self.root = TreeSetNode(item)
        else:
            node = self._find_insert(item)
            
            if item < node.key:
                node.left = TreeSetNode(item)
                node.left.parent = node
                self.root = self._rebalance(node)
            elif item > node.key:
                node.right = TreeSetNode(item)
                node.right.parent = node
                self.root = self._rebalance(node)
    
    
    def insert(self, key):
        """Insert an item.
        
        The function 'add(item)' should be used instead for sets."""
        
        self.add(key)
                
                
    def remove(self, key):
        """Remove a key from the tree.
        
         Raises a KeyError if key is not in the tree."""
        
        node = self._find(key)
        
        if not node:
            raise KeyError(str(key))
            
        self._delete_node(node)
    
    
    def discard(self, key):
        """Remove a key from the tree if present."""
        
        node = self._find(key)
        
        if node:
            self._delete_node(node)
        
    
    def pop(self):
        """Remove and return the greatest item.
        
        Raises an IndexError if the tree is empty."""
         
        if not self.root:
            raise IndexError('pop from empty tree')
        
        node = self._biggest_in_subtree(self.root)
            
        self._delete_node(node)
        return node.key
        
    
    def clear(self):
        """Removes all items from the tree."""
        
        self.root = None
        
    
    def difference_update(self, t):
        """Discard all elements found in t."""
        
        for item in t:
            self.discard(item)
    
    
    def key_at(self, index):
        """Return the key at the index.
        
        Same as '__getitem__(index)' for 'TreeSet' but not for 'TreeDict'."""
        
        return self._node_at(index).key
    
    
    def remove_at(self, index):
        """Remove node at the index."""
        
        self._delete_node(self._node_at(index))
    
    
    def pop_at(self, index):
        """Remove item at the index and return it."""
        
        node = self._node_at(index)
        self._delete_node(node)
        return node.key
    
    
    def _node_at(self, index):
        """Return the node at the index."""
        
        if index < 0:
            if index < -self.root.size:
                raise IndexError('index {} is out of range'.format(index))
            else:
                index += self.root.size
        
        if index >= self.root.size:
            raise IndexError('index {} is out of range'.format(index))
        
        current = self.root
        current_sum = 0
        
        while current:
            current_index = current_sum + current.left_size()
            if index == current_index:
                return current
            elif index < current_index:
                current = current.left
            else:
                current = current.right
                current_sum = current_index + 1
                
        raise RuntimeError('could not find node with index {}'.format(index))
    
    
    def _find(self, key):
        
        if not self.root:
            return None
        
        current = self.root
        while current:
            if key == current.key:
                return current
            elif key < current.key:
                current = current.left
            else:
                current = current.right
                
                
    def _find_insert(self, key):
        
        current = self.root
        while True:
            if key < current.key and current.left:
                current = current.left
            elif key > current.key and current.right:
                current = current.right
            else:
                return current
    
    
    def _delete_node(self, node):
        
        if node.left and node.right:
            
            # replace by smallest in right subtree
            subst = self._smallest_in_subtree(node.right)
            to_rebalance = subst.parent
            if to_rebalance is node:
                to_rebalance = subst
            else:
                to_rebalance.left = subst.right
                if subst.right:
                    subst.right.parent = to_rebalance
                subst.right = node.right
                node.right.parent = subst
            subst.left = node.left
            node.left.parent = subst
            par = node.parent
            subst.parent = par
            if par:
                if node is par.left:
                    par.left = subst
                elif node is par.right:
                    par.right = subst
            else:
                self.root = subst
            self.root = self._rebalance(to_rebalance)
            
        elif node.left:
            par = node.parent
            node.left.parent = par
            if par:
                if node is par.left:
                    par.left = node.left
                elif node is par.right:
                    par.right = node.left
                self.root = self._rebalance(par)
            else:
                self.root = node.left
                
        elif node.right:
            par = node.parent
            node.right.parent = par
            if par:
                if node is par.left:
                    par.left = node.right
                elif node is par.right:
                    par.right = node.right
                self.root = self._rebalance(par)
            else:
                self.root = node.right
                
        else:
            par = node.parent
            if par:
                if node is par.left:
                    par.left = None
                elif node is par.right:
                    par.right = None
                self.root = self._rebalance(par)
            else:
                self.root = None
        
        node.parent, node.left, node.right = None, None, None
    
    
    def _rebalance(self, node):
        
        while node:
            balance = node.update()
            while abs(balance) > 1:
                if balance > 1:
                    if node.left.balance >= 0:
                        # single rotation needed
                        self._rightrotate(node)
                    else:      
                        # double rotation needed
                        self._leftrotate(node.left)
                        self._rightrotate(node)
                else:
                    if node.right.balance <= 0:
                        # single rotation needed
                        self._leftrotate(node)
                    else:
                        # double rotation needed
                        self._rightrotate(node.right)
                        self._leftrotate(node)
                balance = node.balance
            if not node.parent:
                return node
            node = node.parent
            
            
    def _rightrotate(self, y):
        
        x = y.left
        B = x.right
        if y.parent and (y is y.parent.right):
            y.parent.right = x
        elif y.parent and (y is y.parent.left):
            y.parent.left = x
        x.parent = y.parent
        y.parent = x
        x.right = y
        y.left = B
        if B:
            B.parent = y
        y.update()
        x.update()
        
        
    def _leftrotate(self, x):
        
        y = x.right
        B = y.left
        if x.parent and (x is x.parent.right):
            x.parent.right = y
        elif x.parent and (x is x.parent.left):
            x.parent.left = y
        y.parent = x.parent
        x.parent = y
        y.left = x
        x.right = B
        if B:
            B.parent = x
        x.update()
        y.update()
    
    
    def _smallest_in_subtree(self, node):
        
        current = node
        while current.left:
            current = current.left
        return current
    
    
    def _biggest_in_subtree(self, node):
        
        current = node
        while current.right:
            current = current.right
        return current
    
    
    def copy(self):
        """Return a copy of the tree."""
        
        tree_copy = TreeSet()
        if self.root:
            tree_copy.root = self._copy(self.root)
        return tree_copy
    
    
    def _copy(self, node, parent=None):
        
        node_copy = node.copy()
        node_copy.parent = parent
        if node.left:
            node_copy.left = self._copy(node.left, parent=node_copy)
        if node.right:
            node_copy.right = self._copy(node.right, parent=node_copy)
        return node_copy
                
    
    def to_newick(self):
        """Recursive Tree --> Newick (str) function.
        
        Intended for testing purpose.
        """
        
        def _newick(node):
            if not (node.left or node.right):
                return str(node.key)# + "-" + str(node.size)
            else:
                if node.left and node.right:
                    s = '({},{})'.format(_newick(node.left),
                                         _newick(node.right))
                elif node.left:
                    s = '({},-)'.format(_newick(node.left))
                elif node.right:
                    s = '(-,{})'.format(_newick(node.right))
                else:
                    s = ''
                return s + str(node.key)# + "-" + str(node.size)
            
        return _newick(self.root) if self.root else ''
    
    
    def check_integrity(self, node=None):
        """Recursive check of the following properties of AVL trees:
            
        - all children have a correct parent reference
        - the tree is balanced (AVL property)
        - the size is correct in all subtrees
        Intended for testing purpose.
        """
        
        if not node:
            if not self.root:
                print("Tree has no root!")
                return False
            else:
                return self.check_integrity(self.root)
        else:
            height_left, height_right, size = 0, 0, 1
            if node.left:
                if (node is not node.left.parent or 
                    not self.check_integrity(node.left)):
                    print("Check node: '{}' left".format(node))
                    return False
                height_left = node.left.height
                size += node.left.size
            if node.right:
                if (node is not node.right.parent or
                    not self.check_integrity(node.right)):
                    print("Check node:'{}' right".format(node))
                    return False
                height_right = node.right.height
                size += node.right.size
            if abs(node.balance) > 1:
                print("Node '{}' is unbalanced!".format(node))
                return False
            if node.height != 1 + max(height_left, height_right):
                print("Height of node '{}' is incorrect!".format(node))
                return False
            if node.size != size:
                print("Size of node '{}' is incorrect!".format(node))
                return False
            return True
        
        
class TreeSetIterator:
    """Iterator for AVL-tree-based set."""
    
    __slots__ = ('avl_tree', '_current', '_from')
    
    def __init__(self, avl_tree):
        
        self.avl_tree = avl_tree
        self._current = self.avl_tree.root
        
        # Where do I come from?
        # 1 -- up
        # 2 -- left
        # 3 -- right
        self._from = 1
        
        
    def __iter__(self):

        return self
        
    
    def __next__(self):
        
        node = self._find_next()
        if node:
            return node.key
        else:
            raise StopIteration
    
    
    def _find_next(self):
        
        while self._current:
            
            # coming from above
            if self._from == 1:
                if self._current.left:
                    self._current = self._current.left
                else:
                    self._from = 2
            # coming from left child --> return this node
            elif self._from == 2:
                x = self._current
                if self._current.right:
                    self._current = self._current.right
                    self._from = 1
                else:
                    self._current = self._current.parent
                    if self._current and self._current.left is x:
                        self._from = 2
                    elif self._current:
                        self._from = 3
                return x
            # coming from right child
            else:
                x = self._current
                self._current = self._current.parent
                if self._current and self._current.left is x:
                    self._from = 2
                elif self._current:
                    self._from = 3
        

class TreeDictNode(TreeSetNode):
    
    __slots__ = ('value',)
    
    def __init__(self, key, value):
        
        super().__init__(key)
        self.value = value
        
    
    def copy(self):
        
        copy = TreeDictNode(self.key, self.value)
        copy.height = self.height
        copy.size = self.size
        copy.balance = self.balance
        return copy
    

class TreeDict(TreeSet):
        
    
    def __iter__(self):
        
        return TreeDictIterator(self)
        
    
    def __getitem__(self, item):
        
        node = self._find(key)
        
        if not node:
            raise KeyError(str(key))
            
        return node.value
    
    
    def get(self, item, default=None):
        
        node = self._find(key)
        
        if node:
            return node.value
        else:
            return default
        
    
    def keys(self):
        """Return an iterator for the keys."""
        
        return TreeDictIterator(self, mode=1)
        
    
    def values(self):
        """Return an iterator for the values."""
        
        return TreeDictIterator(self, mode=2)
    
    
    def items(self):
        """Return an iterator for (key, value) pairs."""
        
        return TreeDictIterator(self, mode=3)
    
    
    def pop(self, key):
        """Remove a key from the tree and return its value.
        
         Raises a KeyError if key is not in the tree."""
        
        node = self._find(key)
        
        if not node:
            raise KeyError(str(key))
            
        self._delete_node(node)
        return node.value
    
    
    def value_at(self, index):
        """Return the value at the index."""
        
        return self._node_at(index).value
    
    
    def key_and_value_at(self, index):
        """Return the (key, value) pair at the index."""
        
        node = self._node_at(index)
        return (node.key, node.value)
    
    
    def pop_at(self, index):
        """Remove node at the index and return its (key, value) pair."""
        
        node = self._node_at(index)
        self._delete_node(node)
        return (node.key, node.value)
    
    
    def add(self, key, value):
        """Insert a key and value."""
        
        self.insert(key, value)
    
    
    def insert(self, key, value):
        """Insert a key and value."""
        
        if not self.root:
            self.root = TreeDictNode(key, value)
        else:
            node = self._find_insert(key)
            
            if key < node.key:
                node.left = TreeDictNode(key, value)
                node.left.parent = node
                self.root = self._rebalance(node)
            elif key > node.key:
                node.right = TreeDictNode(key, value)
                node.right.parent = node
                self.root = self._rebalance(node)
                
    def copy(self):
        """Return a copy of the tree."""
        
        tree_copy = TreeDict()
        if self.root:
            tree_copy.root = self._copy(self.root)
        return tree_copy
                
                
class TreeDictIterator(TreeSetIterator):
    """Iterator for AVL-tree-based dictionary."""
    
    __slots__ = ('_mode',)
    
    def __init__(self, avl_tree, mode=1):
        
        super().__init__(avl_tree)
        
        # What is returned?
        # 1 -- key
        # 2 -- value
        # 3 -- (key, value)
        self._mode = mode
    
    def __iter__(self):

        return self
    
    def __next__(self):
        
        node = self._find_next()
        if node:
            if self._mode == 1:
                return node.key
            elif self._mode == 2:
                return node.value
            else:
                return (node.key, node.value)
        else:
            raise StopIteration
        

if __name__ == '__main__':
    
    import random, time
    
    keys = [i for i in range(10000)]
    random.shuffle(keys)
    
    t = TreeDict()
    for key in keys:
        t.insert(key, None)
    # print(t.to_newick())
    
    t = t.copy()
    print(len(t))
    print(t.pop_at(-10000))
    print(len(t))
    print(t.key_and_value_at(980))
    
    l = [key for key in t.items()]
    print(l[-5:])
     
    s = set()
    for key in keys:
        s.add(key)
        
    start_time1 = time.time()
    for i in range(500):
        x = random.randint(0, len(t)-1)
        t.pop_at(x)
    end_time1 = time.time()
    print(len(t))
    
    start_time2 = time.time()
    for i in range(500):
        x = random.choice(tuple(s))
        s.remove(x)
    end_time2 = time.time()
    
    print(end_time1 - start_time1, end_time2 - start_time2)


# -*- coding: utf-8 -*-

"""
Graph Tools.
"""

import itertools, random
import numpy as np
import networkx as nx


__author__ = 'David Schaller'


# ----------------------------------------------------------------------------
#                             Adjacency matrix
# ----------------------------------------------------------------------------

def build_adjacency_matrix(G):
    """Return an adjacency matrix."""
    
    # maps node --> row/column index
    index = {node: i for i, node in enumerate(G.nodes())}
    M = np.zeros( (len(index), len(index)), dtype=np.int8)
    
    for x, neighbors in G.adjacency():
        for y in neighbors:
            M[index[x], index[y]] = 1
            
    return M, index


# ----------------------------------------------------------------------------
#                             Graph comparison
# ----------------------------------------------------------------------------

def graphs_equal(G1, G2):
    """Returns whether two NetworkX graphs (directed or undirected) are equal."""
    
    if (not G1.order() == G2.order()) or (not G1.size() == G2.size()):
        return False
    
    if set(G1.nodes()) != set(G2.nodes()):
        return False
    
    for x, y in G1.edges():
        if not G2.has_edge(x, y):
            return False
    
    return True


def is_subgraph(G1, G2):
    """Returns whether G1 is a subgraph of G2."""
    
    if ((isinstance(G1, nx.Graph) and isinstance(G2, nx.DiGraph)) or
        (isinstance(G1, nx.DiGraph) and isinstance(G2, nx.Graph))):
        return False
    
    if G1.order() > G2.order() or G1.size() > G2.size():
        return False
    
    # vertex set is not a subset
    if not set(G1.nodes()) <= set(G2.nodes()):
        return False
    
    for x, y in G1.edges():
        if not G2.has_edge(x, y):
            return False
    
    return True


def symmetric_diff(G1, G2):
    """Returns the number of edges in the symmetric difference."""
    
    set1 = {x for x in G1.nodes()}
    set2 = {x for x in G2.nodes()}
    
    if set1 != set2:
        raise RuntimeError('graphs do not have the same vertex set')
        return
    
    sym_diff_number = 0
    
    if isinstance(G1, nx.DiGraph):
        generator = itertools.permutations(set1, 2)
    else:
        generator = itertools.combinations(set1, 2)
    
    for x, y in generator:
        if G1.has_edge(x, y) and not G2.has_edge(x, y):
            sym_diff_number += 1
        elif not G1.has_edge(x, y) and G2.has_edge(x, y):
            sym_diff_number += 1
    
    return sym_diff_number


def contingency_table(true_graph, graph, as_dict=True):
    """Contingency table for the edge sets of two graphs.
    
    The two graphs must have the same vertex set.
    """
    
    if (true_graph.order() != graph.order() or
        set(true_graph.nodes()) != set(graph.nodes())):
        raise ValueError('graphs must have the same vertex sets')
            
    
    tp, fp, fn = 0, 0, 0
    
    for u, v in graph.edges():
        if true_graph.has_edge(u,v):
            tp += 1
        else:
            fp += 1
    
    for u, v in true_graph.edges():
        if not graph.has_edge(u, v):
            fn += 1
    
    if isinstance(graph, nx.DiGraph):
        tn = (graph.order() * (graph.order()-1)) - (tp + fp + fn)
    else:
        tn = (graph.order() * (graph.order()-1) // 2) - (tp + fp + fn)
    
    if as_dict:
        return {'tp': tp, 'tn': tn, 'fp': fp, 'fn': fn}
    else:
        return tp, tn, fp, fn


def performance(true_graph, graph):
    """Returns order, size, tp, tn, fp, fn, accuracy, precision and recall
    of a (directed or undirected) graph w.r.t. true graph."""
    
    tp, tn, fp, fn = contingency_table(true_graph, graph, as_dict=False)
    
    accuracy = (tp + tn) / (tp + tn + fp + fn) if tp + tn + fp + fn > 0 else float('nan')
    precision = tp / (tp + fp) if tp + fp > 0 else float('nan')
    recall = tp / (tp + fn) if tp + fn > 0 else float('nan')
    
    return (graph.order(), graph.size(),
            tp, tn, fp, fn,
            accuracy, precision, recall)


def false_edges(true_graph, graph):
    """Returns a graph containing fn and a graph containg fp edges."""
    
    if isinstance(true_graph, nx.DiGraph):
        fn_graph = nx.DiGraph()
        fp_graph = nx.DiGraph()
    else:
        fn_graph = nx.Graph()
        fp_graph = nx.Graph()
        
    fn_graph.add_nodes_from(true_graph.nodes(data=True))
    fp_graph.add_nodes_from(true_graph.nodes(data=True))
        
    for u, v in graph.edges():
        if not true_graph.has_edge(u,v):
            fp_graph.add_edge(u, v)
    
    for u, v in true_graph.edges():
        if not graph.has_edge(u, v):
            fn_graph.add_edge(u, v)
        
    return fn_graph, fp_graph


# ----------------------------------------------------------------------------
#                             Graph coloring
# ----------------------------------------------------------------------------

def is_properly_colored(G):
    """Returns whether a (di)graph is properly colored.
    
    Raises a KeyError if, in any edge, a vertex has no 'color' attribute."""
    
    for u, v in G.edges():
        if G.nodes[u]['color'] == G.nodes[v]['color']:
            return False
    
    return True


def sort_by_colors(graph):
    
    color_dict = {}
    
    for v in graph.nodes():
        color = graph.nodes[v]['color']
        
        if color not in color_dict:
            color_dict[color] = [v]
        else:
            color_dict[color].append(v)
    
    return color_dict


def copy_node_attributes(from_graph, to_graph):
    """Copy node label and color from one graph to another."""
    
    for x in from_graph.nodes():
        if to_graph.has_node(x):
            to_graph.nodes[x]['label'] = from_graph.nodes[x]['label']
            to_graph.nodes[x]['color'] = from_graph.nodes[x]['color']


# ----------------------------------------------------------------------------
#                      Graph generation/manipulation
# ----------------------------------------------------------------------------

def random_graph(N, p=0.5):
    """Construct a random graph on N nodes.
    
    Keyword arguments:
        p - probability that an edge xy is inserted
    """
        
    G = nx.Graph()
    G.add_nodes_from(range(1, N+1))
    
    for x, y in itertools.combinations(range(1, N+1), 2):
        if random.random() < p:
            G.add_edge(x, y)
    
    return G


def disturb_graph(graph, insertion_prob, deletion_prob, inplace=False,
                  preserve_properly_colored=True):
    """Randomly insert/delete edges in a graph."""
    
    if not inplace:
        graph = graph.copy()
    
    for x, y in itertools.combinations(graph.nodes, 2):
        
        if (preserve_properly_colored and
            'color' in graph.nodes[x] and 'color' in graph.nodes[y] and
            graph.nodes[x]['color'] == graph.nodes[y]['color']):
            continue
        
        if not graph.has_edge(x, y) and random.random() <= insertion_prob:
            graph.add_edge(x, y)
        elif graph.has_edge(x, y) and random.random() <= deletion_prob:
            graph.remove_edge(x, y)
        
        # done for undirected graphs
        if not isinstance(graph, nx.DiGraph):
            continue
        
        # other direction for digraphs
        if not graph.has_edge(y, x) and random.random() <= insertion_prob:
            graph.add_edge(y, x)
        elif graph.has_edge(y, x) and random.random() <= deletion_prob:
            graph.remove_edge(y, x)
    
    return graph


# ----------------------------------------------------------------------------
#                      Complete multipartite graphs
# ----------------------------------------------------------------------------

def independent_sets(G):
    """Independent sets of a complete multipartite (i.e. a Fitch graph).
    
    Returns a partition of the graph's vertex set that corresponds to its
    set of independent set if the graph is complete multipartite, and
    False otherwise.
    """
    
    # independent sets are equivalent to cliques in the complement graph
    G = nx.complement(G)
    
    ccs = [list(cc) for cc in nx.connected_components(G)]
    
    for cc in ccs:
        for x, y in itertools.combinations(cc, 2):
            if not G.has_edge(x, y):                # not a clique
                return False
    
    return ccs


def complete_multipartite_graph_from_sets(partition):
    """Construct the complete multipartite graphs from the independent sets."""
    
    G = nx.Graph()
    
    for i in range(len(partition)):
        G.add_nodes_from(partition[i])
        
        for j in range(i+1, len(partition)):
            for x, y in itertools.product(partition[i], partition[j]):
                G.add_edge(x, y)
    
    return G


# -*- coding: utf-8 -*-

"""Algorithms for building trees from rooted triple sets."""


import itertools
import networkx as nx

from tralda.datastructures.Tree import Tree, TreeNode
from tralda.datastructures.Partition import Partition


__author__ = 'David Schaller'


def aho_graph(R, L, weighted=False, triple_weights=None):
    """Construct the auxiliary graph (Aho graph) for BUILD.
        
    Edges {a,b} are optionally weighted by the number of occurrences, resp.,
    sum of weights of triples of the form ab|x or ba|x.
    
    Parameters
    ----------
    R : collection of tuples
        A collection of triples.
    L : collection
        A collection of leaf labels.
    weighted : bool, optional
        If True, weight the edges in the resulting graph accordingto the
        number of corresponding triples (and their weights).
        The default is False.
    triple_weights : dict with tuples as keys and float values, optional
        The weights of the triples. The default is None, in which case every
        triple has weight one.
    
    Returns
    -------
    networkx.Graph
        The undirected Aho graph.
    """
    
    G = nx.Graph()
    G.add_nodes_from(L)

    for a, b, c in R:
        if not G.has_edge(a, b):
            G.add_edge(a, b)
        
        if weighted:
            if triple_weights:
                G[a][b]['weight'] = G[a][b].get('weight', 0.0) + \
                                    triple_weights[a, b, c]
            else:
                G[a][b]['weight'] = G[a][b].get('weight', 0.0) + 1.0
    
    return G


def _triple_connect(G, t):
    
    G.add_edge(t[0], t[1])
    G.add_edge(t[0], t[2])
    

def mtt_partition(L, R, F):
    """Construct the auxiliary partition for the MTT algorithm.
    
    Parameters
    ----------
    L : collection
        A collection of leaf labels.
    R : collection of tuples
        A collection of required triples.
    F : collection of tuples
        A collection of forbidden triples.
        
    Returns
    -------
    Partition
        The auxiliary partition for the MTT algorithm.
    networkx.Graph
        A graph representation of the auxiliary partition.
    """
    
    # auxiliary graph initialized as Aho graph
    G = aho_graph(R, L, weighted=False)
    
    # auxiliary partition
    P = Partition(nx.connected_components(G))
    
    if len(P) == 1:
        return P, G
    
    # aux. set of forbidden triples
    S = {t for t in F if P.separated_xy_z(*t)}
    
    # lookup of forb. triples to which u belongs
    L = {u: [] for u in L}
    for t in F:
        for u in t:
            L[u].append(t)
    
    while S:
        t = S.pop()
        _triple_connect(G, t)
        
        # merge returns the smaller of the two merged sets
        smaller_set = P.merge(t[0], t[2])
        
        # update S by traversing the L(u)
        for u in smaller_set:
            for t in L[u]:
                if t in S and not P.separated_xy_z(*t):
                    S.remove(t)
                    _triple_connect(G, t)
                elif t not in S and P.separated_xy_z(*t):
                    S.add(t)
    
    return P, G


class Build:
    """BUILD algorithm.
    
    References
    ----------
    .. [1] A. V. Aho, Y. Sagiv, T. G. Szymanski, and J. D. Ullman.
       Inferring a tree from lowest common ancestors with an application to
       the optimization of relational expressions.
       SIAM Journal on Computing, 10:405–421, 1981.
       doi: 10.1137/0210030.
    """
    
    def __init__(self, R, L, mincut=False, 
                 weighted_mincut=False, triple_weights=None):
        
        self.R = R
        self.L = L
        self.mincut = mincut
        self.weighted_mincut = weighted_mincut
        self.triple_weights = triple_weights
    
    
    def build_tree(self, return_root=False, print_info=False):
        """Build a tree displaying all triples in R if possible.
        
        Parameters
        ----------
        return_root : bool
            If True, return 'TreeNode' instead of 'Tree' instance.
        print_info : bool
            Print information about inconsistencies.
        
        Returns
        -------
        Tree or None
            The BUILD tree on leaf set L displaying all triples in R if the
            triple set R is consistent.
        """
        
        self.cut_value = 0
        self.cut_list = []
        self.print_info = print_info
        
        root = self._aho(self.L, self.R)
        if not root:
            if self.print_info: print('no such tree exists')
            return None
        else:
            return root if return_root else Tree(root)
    
    
    def _aho(self, L, R):
        """Recursive Aho-algorithm."""
        
        # trivial cases: only one or two leaves left in L
        if len(L) == 1:
            leaf = L.pop()
            return TreeNode(label=leaf)
        elif len(L) == 2:
            node = TreeNode()
            for _ in range(2):
                leaf = L.pop()
                node.add_child(TreeNode(label=leaf))
            return node
            
        help_graph = aho_graph(R, L, weighted=self.weighted_mincut,
                                     triple_weights=self.triple_weights)
        conn_comps = self._connected_components(help_graph)
        
        # return False if less than 2 connected components
        if len(conn_comps) <= 1:
            if self.print_info: print('Connected component:\n', conn_comps)
            return False
        
        # otherwise proceed recursively
        node = TreeNode()                   # place new inner node
        for cc in conn_comps:
            Li = set(cc)                    # list/dictionary --> set
            Ri = []
            for t in R:                     # construct triple subset
                if Li.issuperset(t):
                    Ri.append(t)
            Ti = self._aho(Li, Ri)          # recursive call
            if not Ti:
                return False                # raise False to previous call
            else:
                node.add_child(Ti)          # add root of the subtree
   
        return node
    
    
    def _connected_components(self, aho_graph):
        """Determines the connected components of the graph.
        
        And optionally executes a min cut if there is only one component."""
        
        conn_comps = list(nx.connected_components(aho_graph))
        if (not self.mincut) or len(conn_comps) > 1:
            return conn_comps
        else:
            # Stoer–Wagner algorithm
            cut_value, partition = nx.stoer_wagner(aho_graph)
            self.cut_value += cut_value
            if len(partition[0]) < len(partition[1]):
                smaller_comp = partition[0]
            else:
                smaller_comp = partition[1]
            for edge in aho_graph.edges():
                if ((edge[0] in smaller_comp and edge[1] not in smaller_comp)
                    or
                    (edge[1] in smaller_comp and edge[0] not in smaller_comp)):
                    self.cut_list.append(edge)
            return partition
        

class MTT:
    """MTT algorithm.
    
    References
    ----------
    .. [1] Y.-J. He, T. N. D. Huynh, J. Jansson, andW.-K. Sung.
       Inferring phylogenetic relationships avoiding forbidden rooted triplets.
       Journal of Bioinformatics and Computational Biology, 4: 59–74, 2006.
       doi: 10.1142/S0219720006001709.
    """
    
    def __init__(self, R, L, F=None):
        """Constructor for class MTT.
        
        Parameters
        ----------
        L : collection
            A collection of leaf labels.
        R : collection of tuples
            A collection of required triples.
        F : collection of tuples, optional
            A collection of forbidden triples.
        """
        
        self.R = R
        self.L = L
        
        # forbidden triples --> activates MTT if non-empty
        self.F = F
    
    
    def build_tree(self, return_root=False):
        """Build a tree displaying all triples in R if possible.
        
        Parameters
        ----------
        return_root : bool
            If True, return 'TreeNode' instead of 'Tree' instance.
        
        Returns
        -------
        Tree or bool
            A tree on leaf set L displaying all triples in R and none in F,
            if such a tree exists; False otherwise.
        """
        
        self.total_cost = 0
        
        if self.F:
            root = self._mtt(self.L, self.R, self.F)
        else:
            root = self._aho(self.L, self.R)
            
        return root if return_root else Tree(root)
    
    
    def _trivial_case(self, L):
        
        if len(L) == 1:
            leaf = L.pop()
            return TreeNode(label=leaf)
        
        elif len(L) == 2:
            node = TreeNode()
            for _ in range(2):
                leaf = L.pop()
                node.add_child(TreeNode(label=leaf))
            return node
    
    
    def _aho(self, L, R):
        """Recursive Aho algorithm."""
        
        # trivial case: one or two leaves left in L
        if len(L) <= 2:
            return self._trivial_case(L)
            
        aux_graph = aho_graph(R, L)
        partition = list(nx.connected_components(aux_graph))
        
        if len(partition) < 2:
            return False
        
        node = TreeNode()                   # place new inner node
        for s in partition:
            Li, Ri = set(s), []
            for t in R:                     # construct triple subset
                if Li.issuperset(t):
                    Ri.append(t)
            Ti = self._aho(Li, Ri)          # recursive call
            if not Ti:
                return False                # raise False to previous call
            else:
                node.add_child(Ti)          # add roots of the subtrees
   
        return node
    
    
    def _mtt(self, L, R, F):
        """Recursive MTT algorithm."""
        
        # trivial case: one or two leaves left in L
        if len(L) <= 2:
            return self._trivial_case(L)
        
        partition, aux_graph = mtt_partition(L, R, F)
        
        if len(partition) < 2:
            return False
        
        node = TreeNode()                   # place new inner node
        for s in partition:
            Li, Ri, Fi = set(s), [], []
            for Xi, X in ((Ri, R), (Fi, F)):
                for t in X:
                    if Li.issuperset(t):
                        Xi.append(t)
            Ti = self._mtt(Li, Ri, Fi)      # recursive call
            if not Ti:
                return False                # raise False to previous call
            else:
                node.add_child(Ti)          # add roots of the subtrees
   
        return node


def greedy_BUILD(R, L, triple_weights=None, return_root=False):
    """Greedy heuristic for triple consistency.
    
    Add triples one by one and checks consistency via BUILD.
    
    Parameters
    ----------
    R : collection of tuples
        A collection of triples.
    L : collection
        A collection of leaf labels.
    triple_weights : dict, optional
        Weights for the triples; default is None in which case all triples are
        uniformly weighted.
    return_root : bool, optional
        If True, return 'TreeNode' instead of 'Tree' instance.
        The default is False.
    
    Returns
    -------
    Tree or TreeNode
    """
        
    if triple_weights:
        triples = sorted(R,
                         key=lambda triple: triple_weights[triple],
                         reverse=True)
    else:
        triples = R
            
    consistent_triples = []
    root = None
    
    for t in triples:
        consistent_triples.append(t)
        build = Build(consistent_triples, L, mincut=False)
        new_root = build.build_tree(return_root=True)
        if new_root:
            root = new_root
        else:
            consistent_triples.pop()
    
    return root if return_root else Tree(root)
    

def best_pair_merge_first(R, L, triple_weights=None, return_root=False):
    """Wu’s (2004) Best-Pair-Merge-First (BPMF) heuristic.
    
    Modified version by Byrka et al. (2010) and added weights.
    
    Parameters
    ----------
    R : collection of tuples
        A collection of triples.
    L : collection
        A collection of leaf labels.
    triple_weights : dict, optional
        Weights for the triples; default is None in which case all triples are
        uniformly weighted.
    return_root : bool, optional
        If True, return 'TreeNode' instead of 'Tree' instance.
        The default is False.
    
    Returns
    -------
    Tree or TreeNode
    """
    
    # initialization
    nodes = {TreeNode(label=leaf): {leaf} for leaf in L}
    leaf_to_node = {}
    
    for node in nodes:
        leaf_to_node[node.label] = node
    
    # merging
    for i in range(len(L)-1):
        
        score = {(S_i, S_j): 0
                 for S_i, S_j in itertools.combinations(nodes.keys(), 2)}
        
        for x, y, z in R:
            
            w = triple_weights[(x,y,z)] if triple_weights else 1
            
            S_i, S_j, S_k = (leaf_to_node[x],
                             leaf_to_node[y],
                             leaf_to_node[z])
            
            if (S_i is not S_j) and (S_i is not S_k) and (S_j is not S_k):
                
                if (S_i, S_j) in score:
                    score[(S_i, S_j)] += 2 * w
                else:
                    score[(S_j, S_i)] += 2 * w
                    
                if (S_i, S_k) in score:
                    score[(S_i, S_k)] -= w
                else:
                    score[(S_k, S_i)] -= w
                    
                if (S_j, S_k) in score:
                    score[(S_j, S_k)] -= w
                else:
                    score[(S_k, S_j)] -= w
        
        current_max = float('-inf')
        S_i, S_j = None, None
        
        for pair, pair_score in score.items():
            
            if pair_score > current_max:
                current_max = pair_score
                S_i, S_j = pair
        
        # create new node S_k connecting S_i and S_j
        S_k = TreeNode()
        S_k.add_child(S_i)
        S_k.add_child(S_j)
        
        nodes[S_k] = nodes[S_i] | nodes[S_j]    # set union
        for leaf in nodes[S_k]:
            leaf_to_node[leaf] = S_k
        
        del nodes[S_i]
        del nodes[S_j]
    
    if len(nodes) != 1:
        raise RuntimeError('more than 1 node left')
    
    root = next(iter(nodes))
    return root if return_root else Tree(root)


def minimal_identifying_triple_set(tree):
    """Construct a minimal set of triples that identifies the tree.
    
    Parameters
    ----------
    tree : Tree
    
    Yields
    -------
    tuple of three ThreeNode instances
    
    References
    ----------
    - Stefan Grünewald, Mike Steel and M. Shel Swenson. Closure operations in
      phylogenetics. Mathematical Biosciences 208 (2007) 521–537.
      DOI: 10.1016/j.mbs.2006.11.005
    """
    
    # representative leaf for each vertex
    repres = {}
    
    for v in tree.postorder():
        repres[v] = repres[v.children[0]] if v.children else v.label
    
    for u, v in tree.inner_edges():
        W = [c for c in v.children]
        for v2 in u.children:
            if v is not v2:
                for i in range(len(W)-1):
                    yield (repres[W[i]], repres[W[i+1]],
                           repres[v2])
                    

def tree_profile_to_triples(trees):
    """Construct leaf set and reprentative triples from a profile of trees.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    tuple of two sets
        The first set contains all leaf labels that appear in the tree profile.
        The second set contains a representive set of triples.
    """
    
    L = set()
    R = set()
        
    for tree in trees:
        
        L.update(l.label for l in tree.leaves())
        R.update((*sorted(t[:2]), t[2])
                 for t in minimal_identifying_triple_set(tree))
    
    return L, R


def BUILD_supertree(trees):
    """Supertree construction based on the BUILD algorithm.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    Tree or bool
        A supertree for the input trees if existent, False otherwise.
    """
    
    L, R = tree_profile_to_triples(trees)
    
    build = Build(R, L, mincut=False)
    tree = build.build_tree()
    
    return tree if tree else False    
    


# -*- coding: utf-8 -*-

"""
Implementation of a greedy solution for the cluster deletion problem for
cographs, and the complete multipartite graph completion problem.
"""


__author__ = 'David Schaller'


from tralda.datastructures.Tree import Tree
from tralda.cograph.Cograph import to_cotree, complement_cograph
from tralda.tools.GraphTools import complete_multipartite_graph_from_sets


def cluster_deletion(cograph):
    """Cluster deletion for cographs.
    
    Returns a partition of a cograph into disjoint cliques with a minimal
    number of edges between the cliques.
    
    Parameters
    ----------
    cograph : networkx.Graph or Tree
        The cograph for which cluster deletion shall be performed.
    
    Returns
    -------
    list of lists
        A partition where each sublist corresponds to a clique in a solution
        of the cluster deletion problem.
    
    Raises
    ------
    RuntimeError
        If the input is not a valid cograph or cotree.
    
    References
    ----------
    .. [1] Gao Y, Hare DR, Nastos J (2013) The cluster deletion problem for
    cographs. Discrete Math 313(23):2763–2771, DOI 10.1016/j.disc.2013.08.017.
    .. [2] Schaller D, Lafond M, Stadler PF, Wieseke N, Hellmuth M (2020)
    Indirect Identification of Horizontal Gene Transfer. (preprint)
    arXiv:2012.08897
    """
    
    cotree = cograph if isinstance(cograph, Tree) else to_cotree(cograph)
    
    if not cotree:
        raise RuntimeError('not a valid cograph/cotree')
    
    P = {}
    
    for u in cotree.postorder():
        
        P[u] = []
        if not u.children:
            P[u].append([u.label])
        
        elif u.label == 'parallel':
            for v in u.children:
                P[u].extend(P[v])
                
            # naive sorting can be replaced by k-way merge-sort
            P[u].sort(key=len, reverse=True)
        
        elif u.label == 'series':
            for v in u.children:
                for i, Q_i in enumerate(P[v]):
                    if i >= len(P[u]):
                        P[u].append([])
                    P[u][i].extend(Q_i)
        
        else:
            raise RuntimeError('invalid cotree')
    
    return P[cotree.root]


def complete_multipartite_completion(cograph, supply_graph=False):
    """Complete multipartite graph completion for cographs.
    
    Returns a partition of the vertex set corresponding to the (maximal)
    independent sets in an optimal edge completion of the cograph to a
    complete multipartite graph.
    
    Parameters
    ----------
    cograph : networkx.Graph or Tree
        The cograph for which complete multipartite graph completion shall be
        performed.
    supply_graph : bool, optional
        If True, the solution is additionally returned as a NetworkX Graph.
    
    Returns
    -------
    list of lists
        A partition where each sublist corresponds to a (maximal) independent
        set in a solution of the complete multipartite graph completion problem.
    networkx.Graph, optional
        The solution as a graph.
    
    Raises
    ------
    RuntimeError
        If the input is not a valid cograph or cotree.
    """
    
    cotree = cograph if isinstance(cograph, Tree) else to_cotree(cograph)
    
    if not cotree:
        raise RuntimeError('not a valid cograph/cotree')
        
    # complete multipartite graph completion is equivalent to 
    # cluster deletion in the complement cograph
    compl_cotree = complement_cograph(cotree, inplace=False)
    
    # clusters are then equivalent to the maximal independent sets
    independent_sets = cluster_deletion(compl_cotree)
    
    if not supply_graph:
        return independent_sets
    else:
        return (independent_sets,
                complete_multipartite_graph_from_sets(independent_sets))



# -*- coding: utf-8 -*-

"""Fast computation of a common refinement of trees on the same leaf set."""

from collections import deque

from tralda.datastructures.Tree import Tree, TreeNode
from tralda.tools.TreeTools import assert_leaf_sets_equal


__author__ = 'David Schaller'


def linear_common_refinement(trees):
    """Minimal common refinement for a set of trees with the same leaf set.
    
    All input trees must be phylogenetic and have the same set of leaf labels.
    In each tree, the leaves' label attributes must be set and unique.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    Tree or bool
        A common refinement of the input trees if existent, False otherwise.
        
    Raises
    ------
    TypeError
        In case of input instances that are not of type Tree.
    RuntimeError
        If the sequence contains empty trees or the tree do not share the same
        set of leaves.
        
    References
    ----------
    .. [1] D. Schaller, M. Hellmuth, P.F. Stadler (2021)
    A Simple Linear-Time Algorithm for the Common Refinement of Rooted
    Phylogenetic Trees on a Common Leaf Set.
    arXiv:2107.00072 [cs.DS]
    """
    
    cr = LinCR(trees)
    return cr.run()


class LinCR:
    """Minimal common refinement for a set of trees with the same leaf set.
    
    References
    ----------
    .. [1] D. Schaller, M. Hellmuth, P.F. Stadler (2021)
    A Simple Linear-Time Algorithm for the Common Refinement of Rooted
    Phylogenetic Trees on a Common Leaf Set.
    arXiv:2107.00072 [cs.DS]
    """
    
    def __init__(self, trees):
        
        self.L = assert_leaf_sets_equal(trees)
        
        if not self.L:
            raise RuntimeError('trees must have the same leaf labels')
        
        self.trees = trees
        
        
    def run(self):
        
        self._initialize()
        self._build_tree()
        
        if self.T:
            if not self.T.is_phylogenetic():
                return False
            if not self._check_displayed():
                return False
            return self.T
        else:
            return False
    
    
    def _initialize(self):
        """Initialize the lookup tables."""
    
        self.T = None           # resulting tree (candidate)
        
        self._leaf_set_cardinalities()
                                
        self.p = [{} for _ in range(len(self.trees))]
                                # v --> lca_Ti (L(Ti(v)))
        
        self.J = {}             # vertex in resulting tree --> indices of
                                # the trees with corresponding vertex
        self.vi_to_v = {}       # inner vertex in input tree --> 
                                # corresponding vertex in resulting tree
        
        self.Q = deque()        # queue
        self.V = set()          # vertices in resulting tree
        
        self.label_dict = {}
        
        for label in self.L:
            v = TreeNode(label=label)
            self.Q.append(v)
            self.V.add(v)
            self.label_dict[label] = v
            self.J[v] = {i: None for i in range(len(self.trees))}
            self.l[v] = 1 
        
        for i, T_i in enumerate(self.trees):
            for v_i in T_i.leaves():
                v = self.label_dict[v_i.label]
                self.p[i][v] = v_i
                self.J[v][i] = v_i
                self.vi_to_v[v_i] = v
        
    
    def _leaf_set_cardinalities(self):
        """Compute the number of leaves below each vertex."""
        
        self.l = {}
        
        for T_i in self.trees:
            for v in T_i.postorder():
                if v.is_leaf():
                    self.l[v] = 1
                else:
                    self.l[v] = sum(self.l[w] for w in v.children)
    
    
    def _build_tree(self):
        """Constructs the candidate for the common refinement tree."""
        
        self.root = None
        
        while self.Q:
            
            v = self.Q.popleft()
            u = None
            l_min = len(self.L)
            J_u = {}
            
            for i in range(len(self.trees)):
                
                if i in self.J[v]:
                    u2 = self.J[v][i].parent
                else:
                    u2 = self.p[i][v]
                
                if u is None or self.l[u2] < l_min:
                    u = u2
                    l_min = self.l[u2]
                    J_u.clear()
                    J_u[i] = u2
                elif self.l[u2] == l_min:
                    J_u[i] = u2
            
            if self.l[v] < l_min:
                if u in self.vi_to_v:
                    u = self.vi_to_v[u]
                else:
                    u = TreeNode()
                    self.J[u] = J_u
                    for i, u_i in J_u.items():
                        self.vi_to_v[u_i] = u
                        self.J[u][i] = u_i
                    self.l[u] = l_min
                u.add_child(v)
            else:
                return False
            
            if u not in self.V and l_min < len(self.L):
                
                self.Q.append(u)
                self.V.add(u)
                if len(self.V) > 2 * len(self.L) - 2:
                    return False
                
                for i in range(len(self.trees)):
                    if i in self.J[u]:
                        self.p[i][u] = self.J[u][i]
                    elif i in self.J[v]:
                        self.p[i][u] = self.J[v][i].parent
                    else:
                        self.p[i][u] = self.p[i][v]
                        
            elif l_min == len(self.L):
                self.root = u
        
        if not self.root:
            raise RuntimeError('could not determine root')
        
        self.T = Tree(self.root)
    
    
    def _check_displayed(self):
        """Checks whether all input trees are displayed by the constructed
        tree."""
        
        for i, T_i in enumerate(self.trees):
            
            T_copy, v_copy = self.T.copy(mapping=True)
            
            to_contract = [(v_copy[v].parent, v_copy[v])
                           for v in self.T.inner_nodes()
                           if i not in self.J[v] and v.parent]
            
            T_copy.contract(to_contract)
            
            for v_i in T_i.preorder():
                if v_i not in self.vi_to_v:
                    return False
                elif v_i.parent is None:
                    if v_copy[self.vi_to_v[v_i]].parent is not None:
                        return False
                elif v_i.parent not in self.vi_to_v:
                    return False
                elif v_copy[self.vi_to_v[v_i.parent]] is not \
                     v_copy[self.vi_to_v[v_i]].parent:
                    return False
        
        return True



# -*- coding: utf-8 -*-

"""
Euler tour tree as AVL-tree.
    
Implementation of an Euler tour tree for dynamic graph algorithm. A version
of AVL-trees is used as balanced binary tree.

References:
    - Monika Rauch Henzinger, Valerie King. Randomized fully dynamic graph 
      algorithms with polylogarithmic time per operation. J. ACM 46(4) 
      July 1999. 502–536.
"""

from tralda.datastructures.DoublyLinkedList import DLList


__author__ = 'David Schaller'


class DGNode:
    """Dynamic graph node."""
    
    __slots__ = ['value', 'tree_edges', 'nontree_edges',
                 'active_occ', 'occurrences']
    
    def __init__(self, value, active_occ=None):
        
        self.value = value
        self.tree_edges = DLList()
        self.nontree_edges = DLList()
        self.active_occ = active_occ
        
        # list of occurrences (of this node) in the Euler tour
        self.occurrences = DLList()
        if active_occ:
            active_occ.occ = self.occurrences.append(active_occ)


class ETTreeNode:
    
    __slots__ = ('value', 'parent', 'left', 'right', 'prev_occ', 'next_occ',
                 'height', 'size', 'active', 'occ', 'ett')
    
    def __init__(self, value, prev_occ=None, next_occ=None,
                 active=False, ett=None):
        
        self.value = value
        self.parent = None
        self.left = None
        self.right = None
        
        self.prev_occ = prev_occ            # reference to previous occurence
        self.next_occ = next_occ            # reference to next occurence
        
        self.height = 1                     # height of the subtree --> for rebalancing
        self.size = 1 if active else 0      # stores number of active occurences in its subtree
        
        self.active = 1 if active else 0    # active occurrence?
        self.occ = None                     # reference to list entry in occurrences
        
        self.ett = None                     # root stores a reference to the ET tree
   
    
    def get_balance(self):
        if self.left and self.right:
            return self.left.height - self.right.height
        elif self.left:
            return self.left.height
        elif self.right:
            return -(self.right.height)
        else:
            return 0
    
    
    def get_root(self):
        current = self
        while current.parent:
            current = current.parent
        return current
    
    
    def __str__(self):
        return "Occ_" + str(self.value)
 

class ETTree:
    """Euler tour tree.
    
    Represents an Euler tour of a (spanning) tree as a balanced binary tree,
    which is here implemented as an AVL-tree.
    """
    
    __slots__ = ('root', 'nodedict', 'start', 'end', 'current_occ')
    
    def __init__(self, root=None, nodedict=None, start=None, end=None):
        
        self.root = root                            # root of the ET-Tree
        if self.root:
            self.root.ett = self
        if nodedict is not None:
            self.nodedict = nodedict                # hash of active Treenodes
        else:
            self.nodedict = dict()
        self.start = start
        self.end = end
    
    
    def __iter__(self):
        
        self.current_occ = self.start
        return self


    def __next__(self):
        
        if self.current_occ:
            x = self.current_occ
            self.current_occ = self.current_occ.next_occ
            return x
        else:
            raise StopIteration
      
    
    def __len__(self):
        
        if self.root:
            return self.root.size
        else:
            return 0
    
    
    def __contains__(self, item):
        
        if item not in self.nodedict:
            return False
        root = self.nodedict[item].active_occ.get_root()
        return root is self.root
    
    
    def get_size(self):
        
        if self.root:
            return self.root.size
        else:
            return 0
    
    
    def insert(self, value):
        
        if not self.root:
            self.root = ETTreeNode(value)
            return
        
        self._insert(value, self.root)
        self.root.ett = self
        
    
    def _insert(self, value, node):
        
        if value == node.value:
            return
        elif value < node.value:
            if node.left:
                self._insert(value, node.left)
            else:
                node.left = ETTreeNode(value, active=True)
                node.left.parent = node
                self.nodedict[value] = DGNode(value, active_occ=node.left)
                self.root = self.rebalance(node)
        else:
            if node.right:
                self._insert(value, node.right)
            else:
                node.right = ETTreeNode(value, active=True)
                node.right.parent = node
                self.nodedict[value] = DGNode(value, active_occ=node.right)
                self.root = self.rebalance(node)
    
    
    def delete_node(self, node, update_refs=False):
        
        if node.left and node.right:
            subst = self.smallest_in_subtree(node.right)    # replace by smallest
            to_rebalance = subst.parent                     # in right subtree
            if to_rebalance is node:
                to_rebalance = subst
            else:
                to_rebalance.left = subst.right
                if subst.right:
                    subst.right.parent = to_rebalance
                subst.right = node.right
                node.right.parent = subst
            subst.left = node.left
            node.left.parent = subst
            par = node.parent
            subst.parent = par
            if par:
                if node is par.left:
                    par.left = subst
                elif node is par.right:
                    par.right = subst
            else:
                self.root = subst
            self.root = self.rebalance(to_rebalance)
        elif node.left:
            par = node.parent
            node.left.parent = par
            if par:
                if node is par.left:
                    par.left = node.left
                elif node is par.right:
                    par.right = node.left
                self.root = self.rebalance(par)
            else:
                self.root = node.left
        elif node.right:
            par = node.parent
            node.right.parent = par
            if par:
                if node is par.left:
                    par.left = node.right
                elif node is par.right:
                    par.right = node.right
                self.root = self.rebalance(par)
            else:
                self.root = node.right
        else:
            par = node.parent
            if par:
                if node is par.left:
                    par.left = None
                elif node is par.right:
                    par.right = None
                self.root = self.rebalance(par)
            else:
                self.root = None
        self.root.ett = self
        node.parent, node.left, node.right = None, None, None   
        
        # update of prev./next references and active occurrence
        if update_refs:
            if node.prev_occ:                           # update occ. links
                node.prev_occ.next_occ = node.next_occ
            if node.next_occ:
                node.next_occ.prev_occ = node.prev_occ
            if node is self.start:
                self.start = node.next_occ
            if node is self.end:
                self.end = node.prev_occ
            
            active = self.nodedict[node.value]
            active.occurrences.remove_node(node.occ)
            if node.active:                             # was it the active occurrence?
                node.active = 0
                replace = active.occurrences[0]         # (arbitrary) new active node
                if replace:
                    replace.active = 1
                    self.nodedict[node.value].active_occ = replace
                    self.update_height_size(replace, propagate=True)
                else:                                   # no replacement found
                    print("could not find a replacement node "\
                          "for '{}'".format(node))
                    del self.nodedict[node.value]
            
    
    @staticmethod
    def initialize_from_tree(tree, nodedict=None):
        """Initialize the ET tree from an Euler tour of a rooted tree."""
        
        if not tree:
            return
        
        ett = ETTree(nodedict=nodedict)
        previous = None
        for occ in tree.euler_generator():
            if occ in ett.nodedict:
                if ett.nodedict[occ].active_occ:
                    ett_node = ETTreeNode(occ, active=False, prev_occ=previous)
                    ett_node.occ = ett.nodedict[occ].occurrences.append(ett_node)
                else:
                    ett_node = ETTreeNode(occ, active=True, prev_occ=previous)
                    ett.nodedict[occ].active_occ = ett_node
                    ett_node.occ = ett.nodedict[occ].occurrences.append(ett_node)
            else:
                ett_node = ETTreeNode(occ, active=True, prev_occ=previous)
                ett.nodedict[occ] = DGNode(occ, active_occ=ett_node)
            if previous:
                previous.next_occ = ett_node
            else:
                ett.start = ett_node
            previous = ett_node
            ett.rightinsert(ett_node)
        ett.end = previous
        return ett

    
    def rightinsert(self, new_node):
        """Insert new element that is greater than all previous elements."""
        
        if not self.root:
            self.root = new_node
            self.root.ett = self
            return
        current_node = self.root
        while current_node.right:
            current_node = current_node.right
        current_node.right = new_node
        new_node.parent = current_node
        self.root = self.rebalance(current_node)
        self.root.ett = self
        self.end = new_node
    
    
    def rebalance(self, node, stop=None):
        
        while node:
            self.update_height_size(node)
            balance = node.get_balance()
            while abs(balance) > 1:
                if balance > 1:
                    if node.left.get_balance() >= 0:    # single rotation needed
                        self.rightrotate(node)
                    else:                               # double rotation needed
                        self.leftrotate(node.left)
                        self.rightrotate(node)
                else:
                    if node.right.get_balance() <= 0:   # single rotation needed
                        self.leftrotate(node)
                    else:                               # double rotation needed
                        self.rightrotate(node.right)
                        self.leftrotate(node)
                balance = node.get_balance()
            if (not node.parent) or (node.parent is stop):
                return node
            node = node.parent
            
            
    def rightrotate(self, y):
        
        x = y.left
        B = x.right
        if y.parent and (y is y.parent.right):
            y.parent.right = x
        elif y.parent and (y is y.parent.left):
            y.parent.left = x
        x.parent = y.parent
        y.parent = x
        x.right = y
        y.left = B
        if B:
            B.parent = y
        self.update_height_size(y)
        self.update_height_size(x)
        
        
    def leftrotate(self, x):
        
        y = x.right
        B = y.left
        if x.parent and (x is x.parent.right):
            x.parent.right = y
        elif x.parent and (x is x.parent.left):
            x.parent.left = y
        y.parent = x.parent
        x.parent = y
        y.left = x
        x.right = B
        if B:
            B.parent = x
        self.update_height_size(x)
        self.update_height_size(y)
        
    
    def update_height_size(self, node, propagate=False):
        
        if not node:
            return
        if node.left and node.right:
            node.height = 1 + max(node.left.height, node.right.height)
            node.size = node.active + node.left.size + node.right.size
        elif node.left:
            node.height = 1 + node.left.height
            node.size = node.active + node.left.size
        elif node.right:
            node.height = 1 + node.right.height
            node.size = node.active + node.right.size
        else:
            node.height = 1
            node.size = node.active
        if propagate:                       # update the path to the root
            self.update_height_size(node.parent, propagate=True)
    
    
    def get_root(self, value):
        
        if value not in self.nodedict:
            print("could not find node '{}'".format(value))
            return
        return self.nodedict[value].active_occ.get_root()
        
    
    def smallest_in_subtree(self, node):
        
        current = node
        while current.left:
            current = current.left
        return current
    
    
    def biggest_in_subtree(self, node):
        
        current = node
        while current.right:
            current = current.right
        return current
    
    
    # obsolete
    def get_first_smaller(self, node):
        
        first_smaller = None
        if node.left:
            current = node.left
            while current.right:
                current = current.right
            first_smaller = current
        elif node.parent:
            current = node
            while (current.parent and current.parent.left and 
                   current.parent.left == current):
                current = current.parent
            if (current.parent and current.parent.right and 
                current.parent.right == current):
                first_smaller = current.parent
        return first_smaller
    
    
    # obsolete
    def get_first_greater(self, node):
        
        first_greater = None
        if node.right:
            current = node.right
            while current.left:
                current = current.left
            first_greater = current
        elif node.parent:
            current = node
            while (current.parent and current.parent.right and 
                   current.parent.right == current):
                current = current.parent
            if (current.parent and current.parent.left and 
                current.parent.left == current):
                first_greater = current.parent
        return first_greater
    
    
    def smaller(self, node1, node2):
        """Determines whether occurence 'node1' is smaller than 'node2' with
        respect to the Euler tour."""
        
        if node1 is node2:
            return False
        path1, path2 = [node1], [node2]
        while node1.parent:
            path1.append(node1.parent)
            node1 = node1.parent
        while node2.parent:
            path2.append(node2.parent)
            node2 = node2.parent
        if path1[-1] is not path2[-1]:
            return False
        
        for i in range(-2,-min(len(path1),len(path2))-1,-1):
            if path1[i] is not path2[i]:
                if path1[i] is path1[i].parent.left:
                    return True
                else:
                    return False
        if len(path1) < len(path2):
            if path2[-len(path1)-1] is path1[0].left:
                return False
            else: 
                return True
        elif len(path2) < len(path1):
            if path1[-len(path2)-1] is path2[0].left:
                return True
            else:
                return False
    
    
    def lca(self, node1, node2):
        """Finds the last common ancestor of 2 nodes."""
        
        if node1 is node2:
            return node1
        path1, path2 = [node1], [node2]
        while node1.parent:
            path1.append(node1.parent)
            node1 = node1.parent
        while node2.parent:
            path2.append(node2.parent)
            node2 = node2.parent
        if path1[-1] is not path2[-1]:
            return None
        else:
            lca = path1[-1]
        for i in range(-2,-min(len(path1),len(path2))-1,-1):
            if path1[i] is path2[i]:
                lca = path1[i]
            else:
                break
        return lca
    
    
    def cut(self, node1_val, node2_val):
        
        # vertices equal or not in the tree
        if (node1_val == node2_val or (node1_val not in self.nodedict) or 
            (node2_val not in self.nodedict)):
            print("Could not find edge {" + str(node1_val) + "," + 
                  str(node2_val) + "} in the ET tree! (1)")
            return
        
        # which node is further from ET root --> v, the other node is u
        # --> determine u1, u2, v1, v2
        # search in the shorter occurence list as candidate for v
        u, v = self.nodedict[node1_val], self.nodedict[node2_val]
        if len(u.occurrences) < len(v.occurrences):
            v, u = u, v
        # now v.occurrences has less or equal elements
        
        u1, u2, v1, v2 = None, None, None, None
        for occ in v.occurrences:
            if occ.prev_occ and occ.prev_occ.value == u.value:
                u1, v1 = occ.prev_occ, occ
            if occ.next_occ and occ.next_occ.value == u.value:
                v2, u2 = occ, occ.next_occ
            if u1 and u2:
                break
        if not (u1 and u2):                     # search was not successful
            print("Could not find edge {" + str(node1_val) + "," + 
                  str(node2_val) + "} in the ET tree! (2)")
            return
        if u1 is u2 or self.smaller(u2, u1):    # order must be swapped
            u1, v1, v2, u2 = v2, u2, u1, v1
        
        # now u is enclosing v: ... (u1, v1) ... (v2, u2) ...
        # print(u1, v1, v2, u2)
        
        # trivial case: a single node is enclosed (v1 == v2):
        if v1 is v2:
            self.delete_node(v1)                # parameter update_refs is not
            self.update_height_size(v1)         # necessary here (see below)
            spliced_tree = v1
        # splicing-out of a whole interval:
        else:
            lca = self.lca(v1,v2)
            if not lca:
                print("Nodes v1 and v2 are not connencted!")
                return

            # left side handling
            rebalance1 = None
            subtree1 = v1.left
            subtree2 = v1
            if subtree1:
                v1.left = None
                rebalance1 = subtree1
                subtree1.parent = None
            
            # anything smaller between v1 and lca? --> walk from v1 to lca
            if lca is not v1:
                current = v1
                while current.parent and not (current.parent is lca):
                    if current is current.parent.right:
                        smaller = current.parent
                        subtree2 = smaller.right
                        subtree2.parent = None
                        smaller.right = subtree1
                        if subtree1:
                            subtree1.parent = smaller
                        else:
                            rebalance1 = smaller
                        greater = None
                        current = smaller
                        
                        # walk from smaller to lca to find next node
                        # greater than v1 (at least true for lca.left)
                        while current.parent:
                            if current is current.parent.left:
                                greater = current.parent
                                break
                            current = current.parent
                        if not greater:
                            print("Couldn't find node to attach subtree! (2)")
                            return
                        subtree1 = greater.left
                        subtree1.parent = None
                        greater.left = subtree2
                        subtree2.parent = greater
                        if subtree2.parent is lca:
                            break
                        else:
                            current = subtree2
                    else:
                        current = current.parent
            
            # right side handling
            rebalance4 = None
            subtree4 = v2.right
            subtree3 = v2
            if subtree4:
                v2.right = None
                rebalance4 = subtree4
                subtree4.parent = None
            
            # anything greater between v2 and lca? --> walk from v2 to lca
            if lca is not v2:
                current = v2
                while current.parent and not (current.parent is lca):
                    if current is current.parent.left:
                        greater = current.parent
                        subtree3 = greater.left
                        subtree3.parent = None
                        greater.left = subtree4
                        if subtree4:
                            subtree4.parent = greater
                        else:
                            rebalance4 = greater
                        smaller = None
                        current = greater
                        
                        # walk from greater to lca to find next node
                        # smaller than v2 (at least true for lca.right)
                        while current.parent:
                            if current is current.parent.right:
                                smaller = current.parent
                                break
                            current = current.parent
                        if not smaller:
                            print("Couldn't find node to attach subtree! (3)")
                            return
                        subtree4 = smaller.right
                        subtree4.parent = None
                        smaller.right = subtree3
                        subtree3.parent = smaller
                        if subtree3.parent is lca:
                            break
                        else:
                            current = subtree3
                    else:
                        current = current.parent
            
            # reattaching of the subtrees
            attachment_point = lca.parent
            if attachment_point and (attachment_point.left is lca):
                attach_left = True          # remember where to
            else:                           # attach the subtree
                attach_left = False
                
            # inner sequence (to be spliced out)
            lca.parent = None
            if lca is v1:
                spliced_tree = self.rebalance(v2)
            elif lca is v2:
                spliced_tree = self.rebalance(v1)
            else:
                self.rebalance(v2, stop=lca)
                spliced_tree = self.rebalance(v1)
            
            # fragments of outer sequence
            if subtree1 and subtree4:
                attachment_point1 = self.biggest_in_subtree(rebalance1)
                attachment_point1.right = subtree4
                subtree4.parent = attachment_point1
                subtree1 = self.rebalance(rebalance4)  # passes also rebalance1
            elif subtree1:
                subtree1 = self.rebalance(rebalance1)
            elif subtree4:
                subtree1 = self.rebalance(rebalance4)
            
            # re-ligate outer sequence fragments
            if attachment_point:
                if attach_left:
                    attachment_point.left = subtree1
                else:
                    attachment_point.right = subtree1
                if subtree1:
                    subtree1.parent = attachment_point
                self.root = self.rebalance(attachment_point)
                self.root.ett = self
            else:
                self.root = subtree1
                self.root.ett = self
                subtree1.parent = None

        # handling of the obsolete node u2 (possible active node and occ. links
        # are taken care of in function delete_node(x, update_refs=True) )
        self.delete_node(u2, update_refs=True)
        u1.next_occ = v2.next_occ                   # since u2 is already deleted
        if u1.next_occ:                             # might be None
            u1.next_occ.prev_occ = u1
        else:
            self.end = u1
        v1.prev_occ, v2.next_occ = None, None       # disconnect seq. [v1, v2]
        
        new_ett = ETTree(root=spliced_tree, nodedict=self.nodedict, start=v1, end=v2)
        new_ett.root.ett = new_ett
        
        return (self, new_ett)
    
    
    def reroot(self, value):
        """Reroot an Euler tour at a given value."""
        
        if self.get_root(value) is not self.root:
            print("Node ", value, "is not in this ET tree!")
            return
        old1 = self.smallest_in_subtree(self.root)
        
        # case 1: new root is the old root --> do nothing
        if value == old1.value:
            return self
        
        # case 2: rerooting is necessary
        new_root = self.nodedict[value].occurrences[0]
        if not new_root:
            print("Could not find any occurrence of the new root!")
            return
        second_occ = old1.next_occ                  # remember for later (re-ligation)
        self.delete_node(old1, update_refs=True)    # delete first occ. of the old root
        
        # construct two subsequences: 
        # subtree1 (..., new_root1) and subtree2 [new_root1, ...)
        subtree1 = new_root.left
        subtree2 = new_root
        rebalance1 = None
        if subtree1:
            subtree2.left = None
            subtree1.parent = None
            rebalance1 = subtree1
        # anything smaller on the path to the root?
        # --> walk from new_root to root
        current = new_root
        while current.parent:
            if current is current.parent.right:
                smaller = current.parent
                subtree2 = smaller.right
                subtree2.parent = None
                smaller.right = subtree1
                if subtree1:
                    subtree1.parent = smaller
                else:
                    subtree1 = smaller
                    rebalance1 = smaller
                greater = None
                current = smaller
                while current.parent:
                    if current is current.parent.left:
                        greater = current.parent
                        break
                    current = current.parent
                if greater:
                    subtree1 = greater.left
                    subtree1.parent = None
                    greater.left = subtree2
                    subtree2.parent = greater
                    subtree2 = greater
                    current = subtree2
                else:
                    break
            else:
                current = current.parent
        
        # re-ligate: subtree2 + subtree1 + new occ. of the new root
        subtree2 = self.rebalance(new_root)
        active = self.nodedict[new_root.value]
        new_occ = ETTreeNode(active.value)              # new occ. of the root
        new_occ.occ = active.occurrences.append(new_occ)
        if subtree1:
            subtree1 = self.rebalance(rebalance1)
            
            attachment_point1 = self.biggest_in_subtree(subtree1)
            attachment_point1.right = new_occ           # tree links
            new_occ.parent = attachment_point1
            new_occ.prev_occ = attachment_point1        # ET list links
            attachment_point1.next_occ = new_occ
            
            attachment_point2 = self.biggest_in_subtree(subtree2)
            attachment_point2.right = subtree1          # tree links
            subtree1.parent = attachment_point2
            second_occ.prev_occ = attachment_point2     # ET list links
            attachment_point2.next_occ = second_occ
        else:
            attachment_point = self.biggest_in_subtree(subtree2)
            attachment_point.right = new_occ            # tree links
            new_occ.parent = attachment_point
            new_occ.prev_occ = attachment_point         # ET list links
            attachment_point.next_occ = new_occ
            
        new_root.prev_occ = None
        self.root = self.rebalance(new_occ)
        self.root.ett = self
        self.start, self.end = new_root, new_occ
        
        return self
    
    
    @staticmethod
    def link(ett1, ett2, a, b):
        """Link two ET trees on two given nodes.
        
        Links 'ett1' and 'ett2' at node 'a' and 'b', respectively. Therefore,
        the tours are first rerooted and the merged (+ additional occurrence
        of either 'a' or 'b').
        """
        
        if not (isinstance(ett1, ETTree) and isinstance(ett2, ETTree)):
            raise TypeError("Parameters must be of type 'ETTree'!")
        ett1 = ett1.reroot(a)
        ett2 = ett2.reroot(b)
        if not (ett1 and ett2):
            print("Could not find nodes to link in the ET trees!")
            return
        if ett1.root.size < ett2.root.size:
            ett1, ett2 = ett2, ett1          # now ett2 is smaller
        ett1_last = ett1.end
        ett2_last = ett2.end
        ett2_first = ett2.start
        new_occ = ETTreeNode(ett1_last.value, prev_occ=ett2_last)
        new_occ.occ = ett1.nodedict[ett1_last.value].occurrences.append(new_occ)
        
        ett2_last.right = new_occ            # tree links
        new_occ.parent = ett2_last
        ett1_last.right = ett2.root
        ett2.root.parent = ett1_last
        
        ett2_last.next_occ = new_occ         # ET list links
        ett1_last.next_occ = ett2_first
        ett2_first.prev_occ = ett1_last
        
        new_root = ett1.rebalance(new_occ)
        new_ett = ETTree(root=new_root, nodedict=ett1.nodedict, 
                         start=ett1.start, end=new_occ)
        new_root.ett = new_ett
        
        return new_ett
                
    
    def to_newick(self):
        """Recursive Tree --> Newick (str) function.
        
        Intended for testing purpose.
        """
        
        def construct_newick(node):
            if not (node.left or node.right):
                return str(node.value)# + "-" + str(node.size)+ "-" + str(node.active)
            else:
                if node.left and node.right:
                    s = "(" + construct_newick(node.left) + "," + construct_newick(node.right) + ")"
                elif node.left:
                    s = "(" + construct_newick(node.left) + ",-)"
                elif node.right:
                    s = "(-," + construct_newick(node.right) + ")"
                else:
                    s = ""
                return s + str(node.value)# + "-" + str(node.size) + "-" + str(node.active)
            
        return construct_newick(self.root)
    
    
    def ET_to_list(self):
        """Build a list of the Euler tour.
        
        Intended for testing purpose. Should return the same result like
        the function ET_to_list_recursive().
        """
        
        result = []
        for occ in self:
            result.append(occ.value)
        return result
        
        
    def ET_to_list_recursive(self, node=None):
        """Build a list of the Euler tour (recursive).
        
        Intended for testing purpose.
        """
        
        if not node:
            s = self.ET_to_list_recursive(self.root)
        else:
            s = []
            if node.left:
                s += self.ET_to_list_recursive(node.left)
            s += [node.value]
            if node.right:
                s += self.ET_to_list_recursive(node.right)
        return s
    
    
    def check_integrity(self, node=None):
        """Recursive check of the following properties of the ET tree:
            
        - all children have a correct parent reference
        - the tree is balanced (AVL property)
        - the size (nr. of active occurences) is correct in all subtrees
        Intended for testing purpose.
        """
        
        if not node:
            if not self.root:
                print("Tree has no root!")
                return False
            else:
                return self.check_integrity(self.root)
        else:
            height_left, height_right = 0, 0
            size = node.active
            if node.left:
                if (node is not node.left.parent) or (not self.check_integrity(node.left)):
                    print("Check node:", node, "left")
                    return False
                height_left = node.left.height
                size += node.left.size
            if node.right:
                if (node is not node.right.parent) or (not self.check_integrity(node.right)):
                    print("Check node:", node, "right")
                    return False
                height_right = node.right.height
                size += node.right.size
            if abs(node.get_balance()) > 1:
                print("Node", node, "is unbalanced!")
                return False
            if node.height != 1 + max(height_left, height_right):
                print("Height of node", node, "is incorrect!")
                return False
            if node.size != size:
                print("Size of node", node, "is incorrect!")
                return False
            return True



# -*- coding: utf-8 -*-

from tralda.cograph.Cograph import (to_cograph,
                                    to_cotree,
                                    complement_cograph,
                                    random_cotree,
                                    LinearCographDetector)

from tralda.cograph.CographEditor import edit_to_cograph, CographEditor

from tralda.cograph.ClusterDeletion import (cluster_deletion,
                                            complete_multipartite_completion)


# -*- coding: utf-8 -*-

"""
Poly-logarithmic dynamic graph algorithm for Python 3.

Implementation of the poly-logarithmic dynamic graph structure described by
Holm et al. 2001 as Python 3 module. The datastructure uses Euler tour trees
(Henzinger and King 1999) to determine in O(log n) time whether to given
nodes are connected.

Classes in this module:
    - Edge
    - Level
    - HDTGraph

References:
    - Jacob Holm, Kristian de Lichtenberg, and Mikkel Thorup. 
      Poly-logarithmic deterministic fully-dynamic algorithms for 
      connectivity, minimum spanning tree, 2-edge, and biconnectivity. 
      J. ACM, 48(4):723–760, July 2001.
    - Monika Rauch Henzinger, Valerie King. Randomized fully dynamic graph 
      algorithms with polylogarithmic time per operation. J. ACM 46(4) 
      July 1999. 502–536.
"""

from tralda.datastructures.hdtgraph.ETTree import (ETTree,
                                                   ETTreeNode,
                                                   DGNode)
from tralda.datastructures.Tree import Tree


__author__ = 'David Schaller'


class Edge:
    
    __slots__ = ['e', 'level', 'tree_edge', 'dllist_entries']
    
    def __init__(self, e, level=0, tree_edge=False):
        
        self.e = e                      # edge as tuple (u, v)
        self.level = level              # current level of this edge
        self.tree_edge = tree_edge      # is it a tree edge
        self.dllist_entries = None      # reference to doubly-linked-list
                                        # entries (both ends)
        
        
    def __eq__(self, other):
        
        return (self.e == other.e)
    
    
    def __hash__(self):
        
        return hash( self.e )
    
    
    def __repr__(self):
        
        tree = ' tree' if self.tree_edge else ''
        return '{},{}{}'.format(self.e, self.level, tree)
    

class Level:
    
    def __init__(self, index):
        
        self.index = index
        self.forest = set()         # spanning forest (ET trees)
        self.nodedict = dict()      # maps value --> DGNode (of this level)
    
    
    def connected(self, u, v):
        """Determine whether u and v are connected on this level."""
        
        node1, node2 = None, None
        if u in self.nodedict:
            node1 = self.nodedict[u]
        if v in self.nodedict:
            node2 = self.nodedict[v]
        if not (node1 and node2):
            return False
        return node1.active_occ.get_root() is node2.active_occ.get_root()
    
    
    def add_node(self, v):
        """Add a loose node on this level."""
        
        new_etnode = ETTreeNode(v, active=True)
        self.nodedict[v] = DGNode(v, active_occ=new_etnode)
        self.forest.add( ETTree(root=new_etnode, nodedict=self.nodedict, 
                                start=new_etnode, end=new_etnode) )
    
    
    def connect(self, u, v, edge=None):
        """Connect to ETTs by the edge (u,v) on this level.
        
        Keyword argument:
            edge - reference to the edge (u,v), use this if the edge is 
                   a tree edge on this level, default=None
        """
        
        if u not in self.nodedict:
            self.add_node(u)
        if v not in self.nodedict:
            self.add_node(v)
        ett1 = self.nodedict[u].active_occ.get_root().ett
        ett2 = self.nodedict[v].active_occ.get_root().ett
        if ett1 is ett2:
            print("Nodes", u, "and", v, "on level", self.index, "are already connected!")
            return
        new_ett = ETTree.link(ett1, ett2, u, v)
        if not new_ett:
            print("Link operation was not successful!")
            return
        if edge:
            edge.dllist_entries = (self.nodedict[u].tree_edges.append(edge),
                                   self.nodedict[v].tree_edges.append(edge))
        self.forest.remove(ett1)
        self.forest.remove(ett2)
        self.forest.add(new_ett)
        return True
    
    
    def add_nontree_edge(self, e):
        """Add a non-tree edge to its end points on this level."""
        
        u, v = e.e[0], e.e[1]
        if u not in self.nodedict:      # (Can that ever happen?)
            self.add_node(u)
        if v not in self.nodedict:      # (Can that ever happen?)
            self.add_node(v)
        e.dllist_entries = (self.nodedict[u].nontree_edges.append(e),
                            self.nodedict[v].nontree_edges.append(e))
    
    
    def cut(self, u, v):
        """Cut the tree edge (u,v) on this level."""
        
        if not (u in self.nodedict and v in self.nodedict):
            print("Could not find nodes", u, "and", v, "on level", self.index, "!")
            return
        ett1 = self.nodedict[u].active_occ.get_root().ett
        ett2 = self.nodedict[v].active_occ.get_root().ett
        if ett1 is not ett2:
            print("Nodes", u, "and", v, "are not connected on level", self.index, "!")
        result = ett1.cut(u, v)
        if not result:
            print("Nodes", u, "and", v, "on level", self.index, "are not connected!")
            return
        self.forest.add(result[1])      # result[0] is the original instance
                                        # and therefore already in the forest
        return True
        
        
class HDTGraph:
    
    def __init__(self):
        
        self.levels = [Level(0)]
        self.edges = dict()
    
    
    def get_nodes(self):
        """Generator for the nodes in the graph."""
        
        yield from self.levels[0].nodedict.keys()
    
    
    def get_edges(self):
        """Generator for the edges in the graph."""
        
        yield from self.edges.keys()
    
    
    def has_node(self, v):
        """Determine whether a node is in the graph."""
        
        return v in self.levels[0].nodehash
    
    
    def has_edge(self, u, v):
        """Determine whether an edge is in the graph."""
        
        try:
            if u > v:
                u, v = v, u
        except TypeError:
            if id(u) > id(v):
                u, v = v, u
        return (u, v) in self.edges
    
    
    def connected(self, u, v, level=0):
        """Determine in O(log n) whether u and v are connected in the graph."""
        
        return self.levels[level].connected(u, v)
    
    
    def get_component(self, u):
        """Return the connected component (ETT) of a given value."""
        
        if u in self.levels[0].nodedict:
            return self.levels[0].nodedict[u].active_occ.get_root().ett
        else:
            return None
    
    
    def insert_node(self, v):
        """Insert a loose node into the graph."""
        
        if v in self.levels[0].nodedict:
            print("Node", v, "is already in the graph.")
        else:
            self.levels[0].add_node(v)
    
    
    def insert_edge(self, u, v):
        """Insert an edge into the graph."""
        
        try:
            if u > v:
                u, v = v, u
        except TypeError:
            if id(u) > id(v):
                u, v = v, u
        e = (u, v)
        if e in self.edges:    # edge already exists
            return
        
        if self.connected(u, v):
            new_edge = Edge(e, tree_edge=False)             # new edge on level 0
            self.levels[0].add_nontree_edge(new_edge)       # append nontree edge to u and v
        else:
            new_edge = Edge(e, tree_edge=True)              # new edge on level 0
            self.levels[0].connect(u, v, edge=new_edge)     # connect + append tree edge to u and v
        self.edges[e] = new_edge
    
    
    def delete_edge(self, u, v):
        """Delete an edge from the graph."""
        
        try:
            if u > v:
                u, v = v, u
        except TypeError:
            if id(u) > id(v):
                u, v = v, u
        e = (u, v)
        if e not in self.edges:     # edge does not exist
            return
        e = self.edges[e]
        self.edges.pop(e.e)
        l = self.levels[e.level]
        
        # e is not a tree edge --> simply delete e
        if e.tree_edge == False:
            if u in l.nodedict:
                l.nodedict[u].nontree_edges.remove_node(e.dllist_entries[0])
            if v in l.nodedict:
                l.nodedict[v].nontree_edges.remove_node(e.dllist_entries[1])
        
        # e is a tree edge --> search for replacement
        else:
            l.nodedict[u].tree_edges.remove_node(e.dllist_entries[0])
            l.nodedict[v].tree_edges.remove_node(e.dllist_entries[1])
            # remove e on all levels <= l(e)
            for i in range(l.index, -1, -1):
                if not self.levels[i].cut(u, v):
                    print("Something went wrong on level", i)
                    return
            self._replace(u, v, l)
    
    
    def _replace(self, u, v, l):
        """Search for an replacement edge to reconnect u and v on level l."""
        
        ett1 = l.nodedict[u].active_occ.get_root().ett
        ett2 = l.nodedict[v].active_occ.get_root().ett
        if ett1.get_size() <= ett2.get_size():
            smaller_ett = ett1
            root2 = ett2.root
        else:
            smaller_ett = ett2
            root2 = ett1.root
        self._raise_tree_edges(smaller_ett, l)
        found = None
        for occ in smaller_ett:
            if occ.active:
                dgnode = l.nodedict[occ.value]
                while dgnode.nontree_edges:
                    f = dgnode.nontree_edges.popleft()
                    if f.e[0] == dgnode.value:
                        other_end = f.e[1]
                        l.nodedict[other_end].nontree_edges.remove_node(f.dllist_entries[1])
                    else:
                        other_end = f.e[0]
                        l.nodedict[other_end].nontree_edges.remove_node(f.dllist_entries[0])
                    
                    # f is a replacement edge
                    if (other_end in l.nodedict) and (l.nodedict[other_end].active_occ.get_root() is root2):
                        found = f
                        break
                    # f is not a replacement edge --> raise to next level
                    else:
                        f.level += 1
                        self.levels[l.index+1].add_nontree_edge(f)
            if found: break
        if found:
            found.tree_edge = True
            found.dllist_entries = None
            u, v = found.e[0], found.e[1]
            l.connect(u, v, found)
            for i in range(l.index-1, -1, -1):  # reconnect on all lower levels
                self.levels[i].connect(u, v)
        elif l.index > 0:
            self._replace(u, v, self.levels[l.index-1])
            
        
    def _raise_tree_edges(self, ett, l):
        """Raise all tree edges of an ETT to the next level."""
        
        if l.index >= len(self.levels)-1:
            self.levels.append( Level(len(self.levels)) ) # create new level
            
        for occ in ett:
            if occ.active:
                dgnode = l.nodedict[occ.value]
                for tree_edge in dgnode.tree_edges:
                    if tree_edge.level == l.index:
                        tree_edge.level += 1
                        u, v = tree_edge.e[0], tree_edge.e[1]
                        self.levels[l.index+1].connect(u, v, edge=tree_edge)
                        
                # remove all tree edges (of this ET tree) on this level
                dgnode.tree_edges.clear()
    
    
    def _add_tree_edges(self, node):
        
            nodedict = self.levels[0].nodedict
            for child in node.children:
                nodedict[child] = DGNode(child)
                if id(node) < id(child):
                    e = (node, child)
                else:
                    e = (child, node)
                    
                new_edge = Edge(e, level=0, tree_edge=True)
                new_edge.dllist_entries = (nodedict[e[0]].tree_edges.append(new_edge),
                                           nodedict[e[1]].tree_edges.append(new_edge))
                self.edges[e] = new_edge
                self._add_tree_edges(child)
    
    
    def add_loose_tree(self, tree):
        """Add the edges of a tree as undirected edges.
        
        The elements of the 'Tree' instance must not yet be nodes in the graph.
        """
        
        if not isinstance(tree, Tree):
            raise TypeError("instance of type 'Tree' required")
        
        nodedict = self.levels[0].nodedict
        nodedict[tree.root] = DGNode(tree.root)
        self._add_tree_edges(tree.root)
        ett = ETTree.initialize_from_tree(tree, nodedict=nodedict)
        if not ett:
            return
        self.levels[0].forest.add(ett)
        
    
    def is_connected(self):
        """Determine if the graph is connected (on level 0)."""
        
        if len(self.levels[0].forest) == 1:
            for ett in self.levels[0].forest:
                return ett
        else:
            return False
    
    
    def component_iterator(self, representative):
        """Iterator over the connected component of a given value."""
        
        for occ in self.get_component(representative):
            if occ.active:
                yield occ.value
    
    
    def print_ett_forest(self, level=0):
        """Print Euler tour of the spanning forest.
        
        Intended for testing purpose.
        Keyword argument:
            level - level of the ETT spanning forest to be printed,
                    choose "all" for printing all levels, default=0
        """
        
        if level == "all":
            for l in self.levels:
                print('----- Level {} -----'.format(l.index))
                for ett in l.forest:
                    print(ett.ET_to_list(), ett.get_size())
        else:
            for ett in self.levels[level].forest:
                print(ett.ET_to_list(), ett.get_size())
    

if __name__ == '__main__':
    
    graph = HDTGraph()
    
    import random
    insert = []
    for i in range(100):
        a = random.randint(0, 20)
        b = random.randint(0, 20)
        if a != b:
            insert.append( (a, b) )
    
    for e in insert:
        graph.insert_edge(e[0], e[1])

    for e in random.sample(insert, 30):
        print('edge', e[0], e[1])
        graph.delete_edge(e[0], e[1])

    graph.print_ett_forest(level='all')



# -*- coding: utf-8 -*-

from tralda.datastructures.DoublyLinkedList import DLList


__author__ = 'David Schaller'


class Partition:
    """Dynamic partition implementation."""
    
    def __init__(self, iterable):
        
        self.P = DLList()       # partition
        self.lookup = {}        # item --> set in partition
        
        for s in iterable:
            dll_node = self.P.append(set(s))
            for x in s:
                self.lookup[x] = dll_node
                
    
    def __len__(self):
        
        return len(self.P)
    
    
    def __iter__(self):
        
        return PartitionIterator(self)


    def __next__(self):
        
        pass
              
                
    def in_same_set(self, x, y):
        
        return self.lookup[x] is self.lookup[y]
    
    
    def separated_xy_z(self, x, y, z):
        
        return (self.lookup[x] is     self.lookup[y] and
                self.lookup[x] is not self.lookup[z])
    
    
    def merge(self, repr1, repr2):
        
        set1 = self.lookup[repr1]
        set2 = self.lookup[repr2]
        
        if set1 is set2:
            return
        
        # ensure that set1 is smaller
        if len(set2._value) < len(set1._value):
            set1, set2 = set2, set1
        
        for x in set1._value:
            self.lookup[x] = set2
        set2._value |= set1._value
        
        self.P.remove_node(set1)
        
        return set1._value
        

class PartitionIterator:
    """Iterator class for Partition class."""
    
    def __init__(self, partition):
        
        self.partition = partition
        self._current = self.partition.P._first

    def __iter__(self):

        return self

    def __next__(self):
        
        if self._current:
            x = self._current
            self._current = self._current._next
            return x._value
        else:
            raise StopIteration
        
        
    
    


#!/usr/bin/env python3

"""
Construction of a loose consensus tree.
"""


__author__ = 'David Schaller'


from tralda.datastructures.Tree import Tree, TreeNode, LCA
from tralda.tools.TreeTools import assert_leaf_sets_equal


def loose_consensus_tree(trees):
    """Compute the loose consensus tree for trees with the same leaf set.
    
    All input trees must be phylogenetic and have the same set of leaf labels.
    In each tree, the leaves' label attributes must be set and unique.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    Tree
        The loose consensus tree; None if 'trees' is empty.
        
    Raises
    ------
    TypeError
        In case of input instances that are not of type Tree.
    RuntimeError
        If the sequence contains empty trees or the tree do not share the same
        set of leaves.
        
    References
    ----------
    .. [1] J. Jansson, C. Shen, W.-K. Sung.
       Improved Algorithms for Constructing Consensus Trees.
       In: J. ACM 63, 3, Article 28 (June 2016), 24 pages.
       doi:10.1145/2925985.
    """
    
    LC = LooseConsensusTree(trees)
    return LC.run()


class LooseConsensusTree:
    """Construction of the loose consensus tree.
        
    References
    ----------
    .. [1] J. Jansson, C. Shen, W.-K. Sung.
       Improved Algorithms for Constructing Consensus Trees.
       In: J. ACM 63, 3, Article 28 (June 2016), 24 pages.
       doi:10.1145/2925985.
    """
    
    def __init__(self, trees):
        
        if not assert_leaf_sets_equal(trees):
            raise RuntimeError('trees must have the same leaf labels')
        self.trees = trees
    
    
    def run(self):
        
        self.total_contracted = 0
        
        k = len(self.trees)
        if k == 0:
            return None
        elif k == 1:
            return self.trees[0].copy()
        
        T = self.trees[0]
        for j in range(1, k):
            A, x = one_way_compatible(T, self.trees[j],
                                      return_no_of_contractions=True)
            self.total_contracted += x
            T = merge_trees(A, self.trees[j])
        
        for j in range(k):
            T, x = one_way_compatible(T, self.trees[j],
                                      return_no_of_contractions=True)
            self.total_contracted += x
        
        self.T = T
        return self.T


def merge_all(trees):
    """Common refinement for compatible trees with the same leaf set.
    
    All input trees must be phylogenetic and have the same set of leaf labels.
    In each tree, the leaves' label attributes must be set and unique.
    
    Warning: The behavior of the function is undefined if the trees are not
    compatible, i.e., if no common refinement exists.
    
    Parameters
    ----------
    trees : sequence of Tree instances
    
    Returns
    -------
    Tree
        A common refinement of the input trees; None if 'trees' is empty.
    """
    
    if not assert_leaf_sets_equal(trees):
        raise RuntimeError('trees must have the same leaf labels')
    
    k = len(trees)
    if k == 0:
        return None
    elif k == 1:
        return trees[0].copy()
    
    T = trees[0]
    for j in range(1, k):
        T = merge_trees(T, trees[j])
    
    return T


def _preprocess(T1, T2):
    
    # bijection such that clusters in T1 form sequences os consecutive integers
    f = {l.label: i for i, l in enumerate(T1.leaves())}
    
    # dict storing the minimal f(x) for each vertex in T2
    m = {}
    for v in T2.postorder():
        m[v] = min(m[c] for c in v.children) if v.children else f[v.label]
    
    counting_sort = [[] for i in range(len(f))]
    for v in T2.preorder():
        counting_sort[m[v]].append(v)
    sorted_vertices = [v for sublist in counting_sort for v in sublist]
    
    # copy T2 while ordering all children according to counting_sort
    orig_to_new = {}
    for orig in sorted_vertices:
        new = TreeNode()
        orig_to_new[orig] = new
        if orig.parent:
            orig_to_new[orig.parent].add_child(new)
        for key, value in orig.attributes():
            setattr(new, key, value)
    T2 = Tree(orig_to_new[T2.root])
    
    # dict storing for each vertex the number edges from the root
    depth = {}
    for v in T2.preorder():
        depth[v] = depth[v.parent] + 1 if v.parent else 0
    
    L_T2 = list(T2.leaves())
    leaf_rank = {l.label: i for i, l in enumerate(L_T2)}
    start, stop = {}, {}
    for u in T1.postorder():
        if not u.children:
            start[u] = leaf_rank[u.label]
            stop[u] = leaf_rank[u.label]
        else:
            start[u] = min(start[c] for c in u.children)
            stop[u] = max(stop[c] for c in u.children)
    
    # vertex in T2 with smallest depth whose leftmost leaf is x
    x_left, leftmost = {}, {}
    # vertex in T2 with smallest depth whose rightmost leaf is x
    x_right, rightmost = {}, {}
    
    for v in T2.postorder():
        if not v.children:
            x_left[v] = v
            leftmost[v] = v
            x_right[v] = v
            rightmost[v] = v
        else:
            cur_leftmost = leftmost[v.children[0]]
            leftmost[v] = cur_leftmost
            x_left[cur_leftmost] = v
            
            cur_rightmost = rightmost[v.children[-1]]
            rightmost[v] = cur_rightmost
            x_right[cur_rightmost] = v
    
    lca = LCA(T2)
    
    return T2, L_T2, lca, depth, start, stop, x_left, x_right


def merge_trees(T1, T2):
    """Common refinement of two compatible trees with the same leaf set.
    
    Both input trees must be phylogenetic and have the same set of leaf labels.
    In each tree, the leaves' label attributess must be set and unique.
    
    Parameters
    ----------
    T1 : Tree
    T2 : Tree
    
    Returns
    -------
    Tree
        A tree whose clusters are the union of the cluster of T1 and T2.
        
    References
    ----------
    .. [1] J. Jansson, C. Shen, W.-K. Sung.
       Improved Algorithms for Constructing Consensus Trees.
       In: J. ACM 63, 3, Article 28 (June 2016), 24 pages.
       doi:10.1145/2925985.
    """
    
    # T2 gets copied in _preprocess, so input trees remain unchanged
    T2, L_T2, lca, depth, start, stop, x_left, x_right = _preprocess(T1, T2)
    
    for u in T1.postorder():
        
        if not u.children:
            continue
        
        a = L_T2[start[u]]
        b = L_T2[stop[u]]
        r_u = lca(a, b)
        
        leftmost_child = r_u.children[0]
        rightmost_child = r_u.children[-1]
        
        d_u = x_left[a] if depth[x_left[a]] > depth[r_u] else leftmost_child
        e_u = x_right[b] if depth[x_right[b]] > depth[r_u] else rightmost_child
        
        if d_u is leftmost_child and e_u is rightmost_child:
            continue
        
        c = TreeNode()
        depth[c] = depth[r_u] + 1
        if d_u is not leftmost_child:
            x_left[a] = c
        if e_u is not rightmost_child:
            x_right[b] = c
        r_u.add_child_right_of(c, e_u)
        
        c_children = r_u.child_subsequence(d_u, e_u)
        for child in c_children:
            c.add_child(child)
    
    return T2


def one_way_compatible(T1, T2, return_no_of_contractions=False):
    """Remove all clusters in a tree T1 that are incompatible with T2.
    
    Both input trees must be phylogenetic and have the same set of leaf labels.
    In each tree, the leaves' label attributes must be set and unique.
    
    Parameters
    ----------
    T1 : Tree
    T2 : Tree
    
    Returns
    -------
    Tree
        A copy of T1 with all clusters removed (i.e. the corresponding edges
        are contracted) that are incompatible with T2.
        
    References
    ----------
    .. [1] J. Jansson, C. Shen, W.-K. Sung.
       Improved Algorithms for Constructing Consensus Trees.
       In: J. ACM 63, 3, Article 28 (June 2016), 24 pages.
       doi:10.1145/2925985.
    """
    
    # T2 gets copied in _preprocess, so input trees remain unchanged
    T2, L_T2, lca, depth, start, stop, x_left, x_right = _preprocess(T1, T2)
    
    # size of all cluster in T1
    size = {}
    
    # edges uv such that L(T1(v)) is a cluster that is not compatible with T2
    bad_edges = []
    
    for u in T1.postorder():
        
        size[u] = sum(size[c] for c in u.children) if u.children else 1
        
        if not u.children:
            continue
        
        a = L_T2[start[u]]
        b = L_T2[stop[u]]
        r_u = lca(a, b)
        
        leftmost_child = r_u.children[0]
        rightmost_child = r_u.children[-1]
        
        d_u = x_left[a] if depth[x_left[a]] > depth[r_u] else leftmost_child
        e_u = x_right[b] if depth[x_right[b]] > depth[r_u] else rightmost_child
        
        if ((d_u.parent is not r_u) or (e_u.parent is not r_u) or
            (size[u] != stop[u] - start[u] + 1)):
            
            # this should never appear
            if not u.parent:
                raise RuntimeError('vertex {} has no parent'.format(u.parent))
                
            bad_edges.append((u.parent, u))
    
    T1 = T1.contract(bad_edges, inplace=False)
    
    if return_no_of_contractions:
        return T1, len(bad_edges)
    else:
        return T1



# -*- coding: utf-8 -*-

"""
Linked list.

Implementation of a linked list in Python 3. The list enables access
to single list elements in order to modify/delete values in constant time.
"""

import collections


__author__ = 'David Schaller'


class LinkedListNode:
    """Linked list node."""
    
    __slots__ = ('_value', '_next')
    
    def __init__(self, value, next_node=None):
        
        self._value = value
        self._next = next_node
    
    
    def get(self):
        
        return self._value


class LinkedList:
    """Linked list."""
    
    __slots__ = ('_first', '_last', '_count')
    
    def __init__(self, *args):
        
        self._first = None
        self._last = None
        self._count = 0
        
        for arg in args:
            if isinstance(arg, collections.abc.Iterable):
                for item in arg:
                    self.append(item)
            else:
                self.append(item)
        
    
    def __len__(self):
        
        return self._count
    
    
    def __nonzero__(self):
        
        return True if self._count > 0 else False
    
    
    def __iter__(self):
        
        return LinkedListIterator(self)


    def __next__(self):
        
        pass
    
    
    def __getitem__(self, index):
        
        return self.node_at(index)._value
    
    
    def first(self):
        
        return self._first._value
    
    
    def last(self):
        
        return self._last._value
    
    
    def first_node(self):
        
        return self._first
    
    
    def node_at(self, index):
        
        if not isinstance(index, int):
            raise TypeError("index must be of type 'int'")
            
        if index < (-self._count) or index >= self._count:
            raise IndexError('index {} is out of bounds'.format(index))
            
        if index < 0:
            index += self._count
        
        # search from beginning
        node = self._first
        for _ in range(index):
            node = node._next
            
        return node
    
    
    def append(self, value):
        
        new_end = LinkedListNode(value)
        if self._last:
            self._last._next = new_end
        self._last = new_end
        if not self._first:
            self._first = new_end
        self._count += 1
        return new_end
    
    
    def extend(self, iterable):
        
        for value in iterable:
            self.append(value)
            
    
    def concatenate(self, other):
        """Merges two linked lists.
        
        The instance for which this function is called is extended, the other
        list should no longer be used."""
        
        if not isinstance(other, LinkedList):
            raise TypeError("must be of type 'LinkedList'")
            
        if self._last:
            self._last._next = other._first
        else:
            self._first = other._first
        
        if other._last:
            self._last = other._last
            
        self._count += other._count
        
        return self
    
    
    def append_left(self, value):
        
        new_start = LinkedListNode(value, next_node=self._first)
        self._first = new_start
        if not self._last:
            self._last = new_start
        self._count += 1
        return new_start
    
    
    def remove(self, value):
        """Remove an item by value in O(n)."""
        
        node = self._first
        prev_node = None
        
        while node:
            if node._value == value:
                
                if prev_node:
                    prev_node._next = node._next
                else:
                    self._first = node._next
                    
                if not node._next:
                    self._last = prev_node
                
                self._count -= 1
                return
            
            prev_node = node
            node = node._next
            
        raise KeyError('value {} is not in the linked list'.format(value))
    
    
    def insert_right_of(self, node, value):
        """Insert a new item right of a node of the list in O(1)."""
        
        if node is self._last:
            new_node = self.append(value)
            
        else:
            new_node = LinkedListNode(value, next_node=node._next)
            node._next = new_node
            self._count += 1
        
        return new_node
            
    
    def truncate(self, index):
        """Truncate all nodes starting from the specified index."""
        
        if index <= 0:
            self.clear()
        else:
            new_end = self.node_at(index-1)
            self._last = new_end
            if new_end._next:
                new_end._next = None
            self._count = index
            
    
    def truncate_left(self, n):
        """Truncate n nodes on the left side."""
        
        if n >= self._count:
            self.clear()
        else:
            new_start = self.node_at(n)
            self._first = new_start
            self._count -= n
        
    
    # def popright(self):
    #     """Removes the last node of the list and returns its value."""
        
    #     pass
    
    
    def popleft(self):
        """Removes the first node of the list and returns its value."""
        
        if self._first:
            value = self._first._value
            self._first = self._first._next
            self._first._next = None
            self._count -= 1
            return value
        else:
            raise IndexError('pop from empty linked list')
    
    
    def clear(self):
        
        self._first = None
        self._last = None
        self._count = 0
        
    
    def _count_actual(self):
        """Counts the actual number of nodes."""
        
        current = self._first
        counter = 0
        
        while current:
            counter += 1
            current = current._next
            
        return counter
        

class LinkedListIterator:
    """Iterator class for linked list."""
    
    def __init__(self, llist):
        
        self.llist = llist
        self._current = llist._first

    def __iter__(self):
        
        return self
    
    def __next__(self):
        
        if self._current:
            x = self._current
            self._current = self._current._next
            return x._value
        else:
            raise StopIteration


# -*- coding: utf-8 -*-

"""
Cographs and cotrees.  
"""


import itertools, random
from collections import deque

import networkx as nx

from tralda.datastructures.Tree import Tree, TreeNode, LCA


__author__ = 'David Schaller'
    

def to_cograph(cotree):
    """Returns the cograph corresponding to the cotree.
    
    Parameters
    ----------
    cotree : Tree
        A cotree, i.e., a Tree instance with inner vertex labels 'series' and
        'parallel'.
    
    Returns
    -------
    nexworkx.Graph
        The corresponding cograph with the leaf labels as vertices.
    """
    
    leaves = cotree.leaf_dict()
    G = nx.Graph()
    
    for v in leaves[cotree.root]:
        G.add_node(v.label)
    
    for u in cotree.preorder():
        if u.label == 'series':
            for v1, v2 in itertools.combinations(u.children, 2):
                for l1, l2 in itertools.product(leaves[v1], leaves[v2]):
                    G.add_edge(l1.label, l2.label)
    
    return G  
    

def to_cotree(G):
    """Checks if a graph is a cograph and returns its cotree.
    
    Linear O(|V| + |E|) implementation.
    
    Parameters
    ----------
    G : nexworkx.Graph
        A graph.
    
    Returns
    -------
    Tree or bool
        The cotree representation of the graph, or False if it is not a cograph.
    
    References
    ----------
    .. [1] D. G. Corneil, Y. Perl, and L. K. Stewart.
       A Linear Recognition Algorithm for Cographs.
       In: SIAM J. Comput., 14(4), 926–934 (1985).
       doi: 10.1137/0214065
    """
    
    lcd = LinearCographDetector(G)
    return lcd.recognition()
        
    
def complement_cograph(cotree, inplace=False):
    """Returns the cotree of the complement cograph.
    
    Parameters
    ----------
    cotree : Tree
        A cotree, i.e., a Tree instance with inner vertex labels 'series' and
        'parallel'.
    
    Returns
    -------
    Tree
        The cotree of the complement cograph.
    """
    
    tree = cotree if inplace else cotree.copy()
    
    for v in tree.inner_nodes():
        v.label = 'series' if v.label == 'parallel' else 'parallel'
    
    return tree
    
    
def paths_of_length_2(cotree):
    """Generator for all paths of length 2 (edges) in the cograph.
    
    Parameters
    ----------
    cotree : Tree
        A cotree, i.e., a Tree instance with inner vertex labels 'series' and
        'parallel'.
    
    Yields
    ------
    tuple of three TreeNode instance
        All paths of length 2 in the corresponding cograph.
    """
    
    leaves = cotree.leaf_dict()
    lca = LCA(cotree)
    
    for u in cotree.inner_nodes():
        
        if u.label == 'parallel':
            continue
        
        for v1, v2 in itertools.permutations(u.children, 2):
            for t1, t2 in itertools.combinations(leaves[v1], 2):
                if lca(t1, t2).label == 'parallel':
                    for t3 in leaves[v2]:
                        yield t1, t3, t2
    

def random_cotree(N, force_series_root=False):
    """Creates a random cotree.
    
    Parameters
    ----------
    N : int
        The number of leaves in the resulting tree.
    force_series_root : bool
        If True, the cograph of the resulting cotree will be connected,
        otherwise it may be disconnected; the default is False.
    
    Returns
    -------
    Tree
        A cotree, i.e., a Tree instance with inner vertex labels 'series' and
        'parallel'.
    """
    
    cotree = Tree.random_tree(N)
    
    # assign labels ('series', 'parallel')
    for v in cotree.preorder():
        if v.is_leaf():
            continue
        elif v.parent is None:
            if force_series_root:
                v.label = 'series'
            else:
                v.label = 'series' if random.random() < 0.5 else 'parallel'
        else:
            v.label = 'series' if v.parent.label == 'parallel' else 'parallel'
            
    return cotree


class LinearCographDetector:
    """Linear cograph detection and cotree costruction.
    
    References
    ----------
    .. [1] D. G. Corneil, Y. Perl, and L. K. Stewart.
       A Linear Recognition Algorithm for Cographs.
       In: SIAM J. Comput., 14(4), 926–934 (1985).
       doi: 10.1137/0214065
    """
    
    def __init__(self, G):
        """Constructor of LinearCographDetector class.
        
        Parameters
        ----------
        G : networkx.Graph
            A graph.
        """
        
        if not isinstance(G, nx.Graph):
            raise TypeError('not a NetworkX Graph')
        
        self.G = G
        self.V = [v for v in G.nodes()]
        
        self.T = Tree(None)
        self.already_in_T = set()
        self.leaf_map = {}
        self.node_counter = 0
        
        self.marked = set()
        self.m_u_children = {}              # lists of marked and unmarked children
        self.mark_counter = 0
        self.unmark_counter = 0
        self.md = {}
        
        self.error_message = ''
    
    
    def recognition(self):
        """Run cotree construction.
        
        Returns
        -------
        Tree or bool
            The cotree representation of the graph, or False if it is not a
            cograph.
        """
        
        if len(self.V) == 0:
            raise RuntimeError('empty graph in cograph recognition')
            return self.T
        
        elif len(self.V) == 1:
            self.T.root = TreeNode(label=self.V[0])
            self.md[self.T.root] = 0
            return self.T
        
        v1, v2 = self.V[0], self.V[1]
        self.already_in_T.update([v1, v2])
        
        R = TreeNode(label='series')
        self.md[R] = 0
        self.T.root = R
        
        if self.G.has_edge(v1, v2):
            v1_node = TreeNode(label=v1)
            v2_node = TreeNode(label=v2)
            self.md[v1_node] = 0
            self.md[v2_node] = 0
            R.add_child(v1_node)
            R.add_child(v2_node)
            self.node_counter = 3
        else:
            N = TreeNode(label='parallel')
            self.md[N] = 0
            R.add_child(N)
            v1_node = TreeNode(label=v1)
            v2_node = TreeNode(label=v2)
            self.md[v1_node] = 0
            self.md[v2_node] = 0
            N.add_child(v1_node)
            N.add_child(v2_node)
            self.node_counter = 4
            
        self.leaf_map[v1] = v1_node
        self.leaf_map[v2] = v2_node
        
        if len(self.V) == 2:
            self._remove_single_child_root()
            return self.T
        
        for x in self.V[2:]:
            
            # initialization (necessary?)
            self.marked.clear()
            self.m_u_children.clear()
            self.mark_counter = 0
            self.unmark_counter = 0
            self.already_in_T.add(x)        # add x for subsequent iterations
            
            # call procedure _MARK(x)
            self._MARK(x)
            
            # all nodes in T were marked and unmarked
            if self.node_counter == self.unmark_counter:
                R = self.T.root
                x_node = TreeNode(label=x)
                self.md[x_node] = 0
                R.add_child(x_node)
                self.node_counter += 1
                self.leaf_map[x] = x_node
                continue
            # no nodes in T were marked and unmarked
            elif self.mark_counter == 0:
                # d(R)=1
                if len(self.T.root.children) == 1:
                    N = self.T.root.children[0]
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    N.add_child(x_node)
                    self.node_counter += 1
                else:
                    R_old = self.T.root
                    R_new = TreeNode(label='series')
                    self.md[R_new] = 0
                    N = TreeNode(label='parallel')
                    self.md[N] = 0
                    R_new.add_child(N)
                    N.add_child(R_old)
                    self.T.root = R_new
                    
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    N.add_child(x_node)
                    self.node_counter += 3
                self.leaf_map[x] = x_node
                continue
            
            u = self._find_lowest()
            if not u:
                return False
            
            # label(u)=0 and |A|=1
            if u.label == 'parallel' and len(self.m_u_children[u]) == 1:
                w = self.m_u_children[u][0]
                if w.is_leaf():
                    new_node = TreeNode(label='series')
                    self.md[new_node] = 0
                    u.remove_child(w)
                    u.add_child(new_node)
                    new_node.add_child(w)
                    
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    new_node.add_child(x_node)
                    self.node_counter += 2
                else:
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    w.add_child(x_node)
                    self.node_counter += 1 
            
            # label(u)=1 and |B|=1
            elif (u.label == 'series' and 
                  len(u.children) - len(self.m_u_children[u]) == 1):
                set_A = set(self.m_u_children[u])       # auxiliary set bounded by O(deg(x))
                w = None
                for child in u.children:
                    if child not in set_A:
                        w = child
                        break
                if w.is_leaf():
                    new_node = TreeNode(label='parallel')
                    self.md[new_node] = 0
                    u.remove_child(w)
                    u.add_child(new_node)
                    new_node.add_child(w)
                    
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    new_node.add_child(x_node)
                    self.node_counter += 2
                else:
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    w.add_child(x_node)
                    self.node_counter += 1
            
            else:
                y = TreeNode(label=u.label)
                self.md[y] = 0
                for a in self.m_u_children[u]:
                    u.remove_child(a)
                    y.add_child(a)
                    
                if u.label == 'parallel':
                    new_node = TreeNode(label='series')
                    self.md[new_node] = 0
                    u.add_child(new_node)
                    
                    new_node.add_child(y)
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    new_node.add_child(x_node)
                else:
                    par = u.parent
                    if par is not None:             # u was the root of T
                        par.remove_child(u)
                        par.add_child(y)
                    else:
                        self.T.root = y             # y becomes the new root
                    
                    new_node = TreeNode(label='parallel')
                    self.md[new_node] = 0
                    y.add_child(new_node)
                    new_node.add_child(u)
                    x_node = TreeNode(label=x)
                    self.md[x_node] = 0
                    new_node.add_child(x_node)
                self.node_counter += 3
                
            self.leaf_map[x] = x_node
        
        self._remove_single_child_root()
        return self.T
    
    
    def _MARK(self, x):
        
        for v in self.G.neighbors(x):
            if v in self.already_in_T:
                self.marked.add(self.leaf_map[v])
                self.mark_counter += 1
                
        queue = deque(self.marked)
        
        while queue:                        # contains only d(u)=md(u) nodes
            u = queue.popleft()
            self.marked.remove(u)           # unmark u
            self.unmark_counter += 1
            self.md[u] = 0                  # md(u) <- 0
            if u is not self.T.root:
                w = u.parent                # w <- parent(u)
                if w not in self.marked:
                    self.marked.add(w)      # mark w
                    self.mark_counter += 1
                self.md[w] += 1
                if self.md[w] == len(w.children):
                    queue.append(w)
                    
                if w in self.m_u_children:              # append u to list of
                    self.m_u_children[w].appendleft(u)  # marked and unmarked
                else:                                   # children of w
                    self.m_u_children[w] = deque([u])
                    
        if (self.marked and                             # any vertex remained marked
            len(self.T.root.children) == 1 and 
            self.T.root not in self.marked):
            
            self.marked.add(self.T.root)
            self.mark_counter += 1
    
    
    def _find_lowest(self):
        
        R = self.T.root
        y = 'Lambda'
        
        if R not in self.marked:        # R is not marked
            self.error_message = '(iii): R={}'.format(R)
            return False                # G+x is not a cograph (iii)
        else:
            if self.md[R] != len(R.children) - 1:
                y = R
            self.marked.remove(R)
            self.md[R] = 0
            u = w = R
        
        while self.marked:              # while there are mark vertices
            u = self.marked.pop()       # choose a arbitrary marked vertex u
            
            if y != 'Lambda':
                self.error_message = '(i) or (ii): y={}'.format(y)
                return False            # G+x is not a cograph (i) or (ii)
            
            if u.label == 'series':
                if self.md[u] != len(u.children) - 1:
                    y = u
                if u.parent in self.marked:
                    self.error_message = '(i) and (vi): u={}'.format(u)
                    return False        # G+x is not a cograph (i) and (vi)
                else:
                    t = u.parent.parent
            else:
                y = u
                t = u.parent
            self.md[u] = 0           # u was already unmarked above
            
            # check if the u-w path is part of the legitimate alternating path
            while t is not w:
                if t is R:
                    self.error_message = '(iv): t={}'.format(t)
                    return False        # G+x is not a cograph (iv)
                
                if t not in self.marked:
                    self.error_message = '(iii), (v) or (vi): t={}'.format(t)
                    return False        # G+x is not a cograph (iii), (v) or (vi)
                
                if self.md[t] != len(t.children) - 1:
                    self.error_message = '(ii): t={}'.format(t)
                    return False        # G+x is not a cograph (ii)
                
                if t.parent in self.marked:
                    self.error_message = '(i): t={}'.format(t)
                    return False        # G+x is not a cograph (i)
                
                self.marked.remove(t)   # unmark t
                self.md[t] = 0          # reset md(t)
                t = t.parent.parent
                
            w = u                       # rest w for next choice of marked vertex
        
        return u
    
    
    def _remove_single_child_root(self):
        
        if len(self.T.root.children) == 1:
        
            new_root = self.T.root.children[0]
            new_root.detach()
            self.T.root = new_root